import os

import mlflow
from mlflow.utils.autologging_utils import try_mlflow_log
from pydantic import BaseModel
from pydantic.networks import HttpUrl
from typing import Type

from pypads import logger
from pypads.app.injections.base_logger import TrackedObject, LoggerCall
from pypads.app.injections.injection import InjectionLoggerCall, MultiInjectionLogger
from pypads.model.models import TrackedObjectModel, OutputModel, ArtifactMetaModel
from pypads.utils.logging_util import WriteFormats, get_temp_folder
from pypads.utils.util import is_package_available


# utilities
def _to_node_id(wrappee, ref):
    if ref is not None:
        return id(ref)
    else:
        return id(wrappee)


def _to_node_label(wrappee, ref):
    if ref is not None:
        try:
            return str(wrappee) + str(id(ref))
        except Exception as e:
            logger.warning("Couldn't get the representation of wrappee. Fallback to id " + str(id(ref)) + ". " + str(e))
            return str(id(wrappee)) + str(id(ref))
    else:
        try:
            return str(wrappee)
        except Exception as e:
            logger.warning(
                "Couldn't get representation of the wrappee. Fallback to id " + str(id(wrappee)) + ". " + str(e))
            return str(id(wrappee))


def _to_edge_label(wrappee, include_args, args, kwargs):
    if include_args:
        return str(wrappee) + " args: " + str(args) + " kwargs: " + str(kwargs)
    return str(wrappee)


def _step_number(network, label):
    return str(network.number_of_edges()) + ": " + label


class PipelineTO(TrackedObject):
    """
       Tracking object class for execution workflow/ computational graph.
    """

    class PipelineModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/Pipeline"

        network: dict = ...
        pipeline_type: str = ...
        last_tracked: int = ...

        class Config:
            orm_mode = True
            arbitrary_types_allowed = True

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.PipelineModel

    def __init__(self, *args, tracked_by: LoggerCall, network=None, pipeline_type="", last_tracked=None, **kwargs):
        super().__init__(*args, tracked_by=tracked_by, network=network, pipeline_type=pipeline_type,
                         last_tracked=last_tracked, **kwargs)

    def _get_network(self):
        return self.network

    def _set_network(self, network: dict):
        self.network = network

    def _get_last_tracked(self):
        return self.last_tracked

    def _set_last_tracked(self, last_tracked):
        self.last_tracked = last_tracked

    def _get_artifact_path(self, name=None):
        if name is not None:
            return os.path.join(str(id(self)), "pipeline", name)
        else:
            return os.path.join(str(id(self)), "pipeline",)


class PipelineTrackerILF(MultiInjectionLogger):
    """
    Injection logger that tracks multiple calls.
    """
    name = "PipeLineLogger"
    uri = "https://www.padre-lab.eu/onto/pipeline-logger"

    _dependencies = {"networkx"}

    class PipelineTrackerILFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/PipelineILF-Output"

        pipeline: PipelineTO.get_model_cls() = None

        class Config:
            orm_mode = True

    @classmethod
    def output_schema_class(cls):
        return cls.PipelineTrackerILFOutput

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @staticmethod
    def store(pads, *args, **kwargs):
        pipeline_tracker = pads.cache.run_get(pads.cache.run_get("pipeline_tracker"))
        call = pipeline_tracker.get("call")
        output = pipeline_tracker.get("output")
        pipeline = output.pipeline

        from networkx import MultiDiGraph
        network = MultiDiGraph(pipeline._get_network())
        _pipeline_type = pipeline.pipeline_type
        # global network
        if network is not None and len(network.nodes) > 0:
            from networkx import DiGraph
            from networkx.drawing.nx_agraph import to_agraph
            path = os.path.join(pipeline._base_path(), pipeline._get_artifact_path("pypads_pipeline"))
            # try_write_artifact(path, network, WriteFormats.pickle)
            pipeline._store_artifact(network,
                                     ArtifactMetaModel(path=path, description="networkx graph",
                                                       format=WriteFormats.pickle))
            if is_package_available("networkx"):
                base_folder = get_temp_folder()
                folder = base_folder + "pipeline_graph.png"
                if not os.path.exists(base_folder):
                    os.mkdir(base_folder)
                if is_package_available("agraph") and is_package_available("graphviz"):
                    try:
                        if _pipeline_type == "simple":
                            agraph = to_agraph(DiGraph(network))
                        elif _pipeline_type == "grouped":
                            copy = network.copy()
                            edge_groups = {}
                            for edge in copy.edges:
                                plain = copy.get_edge_data(*edge)['plain_label']
                                if plain not in edge_groups:
                                    edge_groups[plain] = []
                                edge_groups[plain].append(edge)

                            for group, edges in edge_groups.items():
                                label = ""
                                suffix = ""
                                base_edge = None
                                for edge in edges:
                                    splits = copy.get_edge_data(*edge)['label'].split(":")
                                    if label == "":
                                        label = splits[0]
                                        suffix = splits[1]
                                        base_edge = edge
                                    else:
                                        label = label + ", " + splits[0]
                                    copy.remove_edge(*edge)
                                label = label + ":" + suffix
                                copy.add_edge(*base_edge, label=label)
                            agraph = to_agraph(copy)
                        elif _pipeline_type == "grouped_no_count":
                            copy = network.copy()
                            edge_groups = {}
                            for edge in copy.edges:
                                plain = copy.get_edge_data(*edge)['plain_label']
                                if plain not in edge_groups:
                                    edge_groups[plain] = edge

                            copy.remove_edges_from(network.edges())
                            for group, edge in edge_groups.items():
                                copy.add_edge(*edge, label=group)
                            agraph = to_agraph(copy)
                        else:
                            agraph = to_agraph(network)
                        agraph.layout('dot')
                        agraph.draw(folder)
                    except ValueError as e:
                        logger.warning("Failed plotting pipeline: " + str(e))
                elif is_package_available("matplotlib"):
                    import matplotlib.pyplot as plt
                    import networkx as nx
                    pos = nx.spring_layout(network)
                    nx.draw(network, pos)
                    nx.draw_networkx_labels(network, pos=pos)
                    nx.draw_networkx_edge_labels(network, pos)
                    plt.savefig(folder)
                if os.path.exists(folder):
                    path = os.path.join(pipeline._base_path(), pipeline._get_artifact_path())
                    try_mlflow_log(mlflow.log_artifact, folder, artifact_path=path)

        call.output = output.store(pipeline_tracker.get("base_path"))
        call.store()

    def __pre__(self, ctx, *args, _logger_call: InjectionLoggerCall, _pypads_pipeline_type="normal",
                _pypads_pipeline_args=False, _logger_output,
                **kwargs):
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()
        pads.cache.run_add("pipeline_tracker", id(self))

        if _logger_output.pipeline is None:
            pipeline = PipelineTO(tracked_by=_logger_call, pipeline_type=_pypads_pipeline_type)
        else:
            pipeline = _logger_output.pipeline

        network = pipeline.network
        last_tracked = pipeline.last_tracked
        _pipeline_type = pipeline.pipeline_type

        import networkx as nx
        if network is None:
            network = nx.MultiDiGraph()
        else:
            network = nx.MultiDiGraph(network)

        node_id = _to_node_id(_logger_call.original_call.call_id.wrappee, ctx)
        label = _to_node_label(_logger_call.original_call.call_id.wrappee, ctx)
        if not network.has_node(node_id):
            network.add_node(node_id, label=label)

        label = _to_edge_label(_logger_call.original_call.call_id.wrappee, _pypads_pipeline_args, args, kwargs)
        # If the current stack holds only the call itself
        if pads.call_tracker.call_depth() == 1:

            # If there where no nodes until now
            if len(network.nodes) == 1:
                network.add_node(-1, label="entry")
                last_tracked = -1
                pipeline._set_last_tracked(last_tracked)
            network.add_edge(last_tracked, node_id, plain_label=label, label=_step_number(network, label))

        # If the tracked function was called from another tracked function
        elif pads.call_tracker.call_depth() > 1:

            containing_node_id = _to_node_id(pads.call_tracker.call_stack[-2].call_id.wrappee, ctx)
            if not network.has_node(containing_node_id):
                containing_node_label = _to_node_label(pads.call_tracker.call_stack[-2].call_id.wrappee, ctx)
                network.add_node(containing_node_id, label=containing_node_label)
            # Add an edge from the tracked function to the current function call
            network.add_edge(containing_node_id, node_id, plain_label=label, label=_step_number(network, label))

        pipeline._set_network(nx.to_dict_of_dicts(network))
        pipeline.store(_logger_output, "pipeline")
        return node_id

    def __post__(self, ctx, *args, _pypads_pipeline_args=False, _logger_call: InjectionLoggerCall, _logger_output,
                 _pypads_pre_return,
                 **kwargs):
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()

        import networkx as nx
        pipeline = _logger_output.pipeline
        network = nx.MultiDiGraph(pipeline._get_network())

        node_id = _pypads_pre_return
        label = "return " + _to_edge_label(_logger_call.original_call.call_id.wrappee, _pypads_pipeline_args, args,
                                           kwargs)
        if pads.call_tracker.call_depth() == 1:
            network.add_edge(node_id, -1, plain_label=label, label=_step_number(network, label))
        elif pads.call_tracker.call_depth() > 1:
            containing_node_id = _to_node_id(pads.call_tracker.call_stack[-2].call_id.wrappee, ctx)
            if not network.has_node(containing_node_id):
                containing_node_label = _to_node_label(pads.call_tracker.call_stack[-2].call_id.wrappee, ctx)
                network.add_node(containing_node_id, label=containing_node_label)
            network.add_edge(node_id, containing_node_id, plain_label=label, label=_step_number(network, label))
        pipeline._set_network(nx.to_dict_of_dicts(network))
        pipeline.store(_logger_output, "pipeline")



from abc import ABCMeta
from functools import wraps
from typing import List

from pypads.app.misc.extensions import ExtendableMixin, Plugin
from pypads.app.misc.mixins import FunctionHolderMixin
from pypads.importext.mappings import Mapping
from pypads.utils.util import inheritors, get_class_that_defined_method

decorator_plugins = set()


class Decorator(FunctionHolderMixin, metaclass=ABCMeta):

    def __init__(self, *args, fn, **kwargs):
        super().__init__(*args, fn=fn, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.__real_call__(*args, **kwargs)


class IDecorators(Plugin):

    def __init__(self):
        super().__init__(type=Decorator)
        decorator_plugins.add(self)

    def _get_meta(self):
        """ Method returning information about where the actuator was defined."""
        return self.__module__

    def _get_methods(self):
        return [method_name for method_name in dir(self) if callable(getattr(object, method_name))]


def decorator(f):
    """
    Decorator used to convert a function to a tracked actuator.
    :param f:
    :return:
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        # self is an instance of the class
        return Decorator(fn=f)(self, *args, **kwargs)

    return wrapper


class PyPadsDecorators(IDecorators):
    def __init__(self):
        super().__init__()

    @property
    def pypads(self):
        from pypads.app.pypads import get_current_pads
        return get_current_pads()

    @decorator
    def track(self, event="pypads_log", mapping: Mapping = None):
        def track_decorator(fn):
            ctx = get_class_that_defined_method(fn)
            events = event if isinstance(event, List) else [event]
            return self.pypads.api.track(ctx=ctx, fn=fn, anchors=events, mapping=mapping)

        return track_decorator


class DecoratorPluginManager(ExtendableMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(plugin_list=decorator_plugins)


pypads_decorators = PyPadsDecorators()


def decorators():
    """
    Returns classes of
    :return:
    """
    return inheritors(Decorator)



from pypads import logger

from pypads.app.injections.injection import InjectionLogger


class Log(InjectionLogger):
    """
    Function just logging the execution into debug.
    """

    def __pre__(self, ctx, *args, _logger_call, **kwargs):
        logger.debug("Entered " + str(_logger_call.original_call))

    def __post__(self, ctx, *args, _logger_call, **kwargs):
        logger.debug("Exited " + str(_logger_call.original_call))


class LogInit(InjectionLogger):
    """
    This function only writes an information of a constructor execution to the stdout.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __pre__(self, ctx, *args, **kwargs):
        logger.info("Pypads tracked class " + str(ctx.__class__) + " initialized.")



import os
from abc import ABCMeta
from contextlib import contextmanager
from functools import wraps
from typing import List, Iterable

import mlflow
from mlflow.entities import ViewType

from app.env import LoggerEnv
from pypads import logger
from pypads.app.injections.run_loggers import RunSetup, RunTeardown
from pypads.app.misc.caches import Cache
from pypads.app.misc.extensions import ExtendableMixin, Plugin
from pypads.app.misc.mixins import FunctionHolderMixin
from pypads.bindings.anchors import get_anchor, Anchor
from pypads.importext.mappings import Mapping, MatchedMapping, make_run_time_mapping_collection
from pypads.importext.package_path import PackagePathMatcher, PackagePath
from pypads.model.models import TagMetaModel, ParameterMetaModel, MetricMetaModel, MetadataModel, ArtifactMetaModel
from pypads.utils.files_util import get_artifacts
from pypads.utils.logging_util import WriteFormats, try_write_artifact, try_read_artifact, get_temp_folder, \
    _to_artifact_meta_name, _to_metric_meta_name, _to_param_meta_name
from pypads.utils.util import inheritors

api_plugins = set()


class Cmd(FunctionHolderMixin, metaclass=ABCMeta):

    def __init__(self, *args, fn, **kwargs):
        super().__init__(*args, fn=fn, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.__real_call__(*args, **kwargs)


class IApi(Plugin):

    def __init__(self):
        super().__init__(type=Cmd)
        api_plugins.add(self)

    def _get_meta(self):
        """ Method returning information about where the actuator was defined."""
        return self.__module__

    def _get_methods(self):
        return [method_name for method_name in dir(self) if callable(getattr(object, method_name))]


def cmd(f):
    """
    Decorator used to convert a function to a tracked actuator.
    :param f:
    :return:
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        # self is an instance of the class
        return Cmd(fn=f)(self, *args, **kwargs)

    return wrapper


class PyPadsApi(IApi):
    """
    Default api functions of pypads
    """

    def __init__(self):
        super().__init__()

    @property
    def pypads(self):
        from pypads.app.pypads import get_current_pads
        return get_current_pads()

    # noinspection PyMethodMayBeStatic
    @cmd
    def track(self, fn, ctx=None, anchors: List = None, mapping: Mapping = None):
        """
        Method to inject logging capabilities into a function
        :param fn: Function to extend
        :param ctx: Ctx which defined the function
        :param anchors: Anchors to trigger on function call
        :param mapping: Mapping defining this extension
        :return: The extended function
        """

        # Warn if ctx doesn't defined the function we want to track
        if ctx is not None and not hasattr(ctx, fn.__name__):
            logger.warning("Given context " + str(ctx) + " doesn't define " + str(fn.__name__))
            ctx = None

        # If we don't have a valid ctx the fn is unbound, otherwise we can extract the ctx path
        if ctx is not None:
            if hasattr(ctx, '__module__') and ctx.__module__ is not str.__class__.__module__:
                ctx_path = ctx.__module__.__name__
            else:
                ctx_path = ctx.__name__
        else:
            ctx_path = "<unbound>"

        if anchors is None:
            anchors = [get_anchor("pypads_log")]
        elif not isinstance(anchors, Iterable):
            anchors = [anchors]

        _anchors = set()
        for a in anchors:
            if isinstance(a, str):
                anchor = get_anchor(a)
                if anchor is None:
                    anchor = Anchor(a, "No description available")
                _anchors.add(anchor)
            elif isinstance(a, Anchor):
                _anchors.add(a)

        # If no mapping was given a default mapping has to be created
        if mapping is None:
            logger.warning("Tracking a function without a mapping definition. A default mapping will be generated.")
            if '__file__' in fn.__globals__:
                lib = fn.__globals__['__file__']
            else:
                lib = fn.__module__

            # For all events we want to hook to
            mapping = Mapping(PackagePathMatcher(ctx_path + "." + fn.__name__), make_run_time_mapping_collection(lib),
                              _anchors, {"concept": fn.__name__})

        # Wrap the function of given context and return it
        return self.pypads.wrap_manager.wrap(fn, ctx=ctx, matched_mappings={MatchedMapping(mapping, PackagePath(
            ctx_path + "." + fn.__name__))})

    @cmd
    def start_run(self, run_id=None, experiment_id=None, run_name=None, nested=False, _pypads_env=None):
        """
        Method to start a new mlflow run. And run its setup functions.
        Every run is supposed to be an own execution. This may make sense if a single file defines multiple executions
        you want to track. (Entry to hyperparameter searches?)
        :param run_id: The id the run should get. This will be chosen automatically if None.
        :param experiment_id: The id the parent experiment has.
        :param run_name: A name for the run. This will also be chosen automatically if None.
        :param nested: If the run should be a nested run. (Run spawned in context of another run)
        :param _pypads_env: Pass the logging env if one is set.
        :return: The newly spawned run
        """
        out = mlflow.start_run(run_id=run_id, experiment_id=experiment_id, run_name=run_name, nested=nested)
        self.run_setups(
            _pypads_env=_pypads_env or LoggerEnv(parameter=dict(), experiment_id=experiment_id, run_id=run_id))
        return out

    # ---- logging ----
    @cmd
    def log_artifact(self, local_path, meta=None, artifact_path=None):
        """
        Function to log an artifact on local disk. This artifact is transferred into the context of mlflow.
        The context might be a local repository, sftp etc.
        :param artifact_path: Path where to store the artifact
        :param local_path: Path of the artifact to log
        :param meta: Meta information you want to store about the artifact. This is an extension by pypads creating a
        json containing some meta information.
        :return:
        """
        self.pypads.backend.log_artifact(local_path, meta=meta, artifact_path=artifact_path)
        self.log_artifact_meta(os.path.basename(local_path), meta)

    @cmd
    def log_mem_artifact(self, name, obj, write_format=WriteFormats.text.name, path=None, meta=None,
                         preserve_folder=True):
        """
        See log_artifact. This logs directly from memory by storing the memory to a temporary file.
        :param name: Name of the new file to create.
        :param obj: Object you want to store
        :param write_format: Format to write to. WriteFormats currently include text and binary.
        :param meta: Meta information you want to store about the artifact. This is an extension by pypads creating a
        json containing some meta information.
        :param path: Path to which to save this artifact.
        :param preserve_folder: Preserve the folder structure
        :return:
        """
        from pypads.app.pypads import get_current_pads
        from pypads.utils.logging_util import WriteFormats, get_run_folder
        import json

        if path:
            name = os.path.join(path, name)
        if meta is None:
            meta = ArtifactMetaModel(path=name, description='Artifact meta information', format=write_format)
        self.pypads.backend.log_mem_artifact(obj, meta=meta, preserve_folder=preserve_folder)
        if write_format == WriteFormats.json:
            pads = get_current_pads()
            consolidated_dict = pads.cache.get('consolidated_dict', None)
            if consolidated_dict is not None:
                path = os.path.join(get_run_folder(), 'artifact', name)
                if isinstance(obj, str):
                    data = json.loads(obj)
                elif isinstance(obj, dict):
                    data = obj
                else:
                    data = str(write_format)
                consolidated_dict[path] = data
                pads.cache.add('consolidated_dict', consolidated_dict)
        self.log_artifact_meta(name, meta)

    @cmd
    def log_artifact_meta(self, name, meta=None):
        self._write_meta(_to_artifact_meta_name(name), meta)

    @cmd
    def log_metric(self, key, value, step=None, meta: MetricMetaModel = None):
        """
        Log a metric to mlflow.
        :param key: Metric key
        :param value: Metric value
        :param step: A step for metrics which can change while executing
        :param meta: Meta information you want to store about the metric. This is an extension by pypads creating a
        json containing some meta information.
        :return:
        """
        if not meta:
            meta = MetricMetaModel(name=key, step=step, description="Metric meta information")
        self.pypads.backend.log_metric(value, meta=meta)
        self.log_metric_meta(meta.name, meta)

    @cmd
    def log_metric_meta(self, key, meta=None):
        self._write_meta(_to_metric_meta_name(key), meta)

    @cmd
    def log_param(self, key, value, meta: ParameterMetaModel = None):
        """
        Log a parameter of the execution.
        :param key: Parameter key
        :param value: Parameter value
        :param meta: Meta information you want to store about the parameter. This is an extension by pypads creating a
        json containing some meta information.
        :return:
        """
        if not meta:
            meta = ParameterMetaModel(name=key, type=str(type(value)), description="Parameter meta information")
        self.pypads.backend.log_parameter(value, meta=meta)
        self.log_param_meta(meta.name, meta)

    @cmd
    def log_param_meta(self, key, meta=None):
        self._write_meta(_to_param_meta_name(key), meta)

    @cmd
    def set_tag(self, key, value, description="No description given."):
        """
        Set a tag for your current run.
        :param description: Description what this tag indicates
        :param key: Tag key
        :param value: Tag value
        :return:
        """
        return self.pypads.backend.set_tag(value, TagMetaModel(name=key, description=description))

    @cmd
    def store_tracked_object(self, to):
        return self.pypads.backend.store_tracked_object(to=to)

    @cmd
    def store_logger_output(self, lo, path=""):
        return self.pypads.backend.store_logger_output(lo=lo, path=path)

    @cmd
    def write_data_item(self, path, content_item, data_format=None, preserve_folder=True):
        """
        Function to log a tracking object content item on local disk. This artifact is transferred into the context of mlflow.
        The context might be a local repository, sftp etc.
        :param path: path to where to store the content item of a tracking object.
        :param content_item: the content item of a tracking object.
        :param content_format: write format for the object content
        :param preserve_folder:  Preserve the folder structure
        :return:
        """
        try_write_artifact(path, content_item, write_format=data_format, preserve_folder=preserve_folder)

    def _write_meta(self, name, meta, write_format=WriteFormats.yaml):
        """
        Write the meta information about an given object name as artifact.
        :param name: Name of the object
        :param meta: Metainformation to store
        :return:
        """
        if meta:
            self.pypads.backend.log_mem_artifact(meta.json(), MetadataModel(path=name + ".meta",
                                                                 description="Meta information of artifact '{}'".format(
                                                                     name), format=write_format))

    def _read_meta(self, name):
        """
        Read the metainformation of a object name.
        :param name:
        :return:
        """
        # TODO format / json / etc?
        return try_read_artifact(name + ".meta.yaml")

    @cmd
    def artifact(self, name):
        return try_read_artifact(name)

    @cmd
    def metric_meta(self, name):
        """
        Load the meta information of a metric by given name.
        :param name: Name of the metric
        :return:
        """
        return self._read_meta(_to_metric_meta_name(name))

    @cmd
    def param_meta(self, name):
        """
        Load the meta information of a parameter by given name.
        :param name: Name of the parameter
        :return:
        """
        return self._read_meta(_to_param_meta_name(name))

    @cmd
    def artifact_meta(self, name):
        """
        Load the meta information of an artifact by given name.
        :param name: Name of the artifact
        :return:
        """
        return self._read_meta(_to_artifact_meta_name(name))

    # !--- logging ----

    # ---- run management ----
    @contextmanager
    @cmd
    def intermediate_run(self, **kwargs):
        """
        Spawn an intermediate nested run.
        This run closes automatically after the "with" block and restarts the parent run.
        :param kwargs: Other kwargs to pass to start_run()
        :return:
        """
        enclosing_run = mlflow.active_run()
        try:
            run = self.pypads.api.start_run(**kwargs, nested=True)
            self.pypads.cache.run_add("enclosing_run", enclosing_run)
            yield run
        finally:
            if not mlflow.active_run() is enclosing_run:
                self.pypads.api.end_run()
                self.pypads.cache.run_clear()
                self.pypads.cache.run_delete()
            else:
                mlflow.start_run(run_id=enclosing_run.info.run_id)

    def _get_setup_cache(self):
        """
        Get registered pre_run functions.
        :return:
        """
        if not self.pypads.cache.exists("pre_run_fns"):
            pre_run_fn_cache = Cache()
            self.pypads.cache.add("pre_run_fns", pre_run_fn_cache)
        return self.pypads.cache.get("pre_run_fns")

    @cmd
    def register_setup(self, name, pre_fn: RunSetup, silent=True):
        """
        Register a new pre_run function.
        :param name: Name of the registration
        :param pre_fn: Function to register
        :param silent: Ignore log output if pre_run was already registered.
        This makes sense if a logger running multiple times wants to register a single setup function.
        :return:
        """
        cache = self._get_setup_cache()
        if cache.exists(name):
            if not silent:
                logger.debug("Pre run fn with name '" + name + "' already exists. Skipped.")
        else:
            cache.add(name, pre_fn)

    @cmd
    def register_setup_fn(self, name, description, fn, nested=True, intermediate=True, order=0, silent=True):
        """
        Register a new pre_run_function by building it from given parameters.
        :param description: A description of the setup function.
        :param name: Name of the registration
        :param fn: Function to register
        :param nested: Parameter if this function should be called on nested runs.
        :param intermediate: Parameter if this function should be called on a intermediate run.
        An intermediate run is a nested run managed specifically by pypads.
        :param order: Value defining the execution order for pre run function.
        The lower the value the sooner a function gets executed.
        :param silent:
        :return:
        """

        class TmpRunSetupFunction(RunSetup):
            pass

        TmpRunSetupFunction.__doc__ = description

        self.register_setup(name, TmpRunSetupFunction(fn=fn, nested=nested, intermediate=intermediate, order=order),
                            silent=silent)

    @cmd
    def run_setups(self, _pypads_env=None):
        cache = self._get_setup_cache()
        fns = []
        for k, v in cache.items():
            fns.append(v)
        fns.sort(key=lambda f: f.order)
        for fn in fns:
            if callable(fn):
                fn(self, _pypads_env=_pypads_env)

    def _get_teardown_cache(self):
        """
        Get the teardown function registry from cache
        :return:
        """
        # General post run cache
        if not self.pypads.api.active_run():
            if not self.pypads.cache.exists("post_run_fns"):
                post_run_fn_cache = Cache()
                self.pypads.cache.add("post_run_fns", post_run_fn_cache)
            return self.pypads.cache.get("post_run_fns")

        # Post run cache for especially this run
        if not self.pypads.cache.run_exists("post_run_fns"):
            post_run_fn_cache = Cache()
            self.pypads.cache.run_add("post_run_fns", post_run_fn_cache)
        return self.pypads.cache.run_get("post_run_fns")

    @cmd
    def register_teardown(self, name, post_fn: RunTeardown, silent=True):
        """
        Register a new post run function.
        :param name: Name of the registration
        :param post_fn: Function to register
        :param silent: Ignore log output if post_run was already registered.
        This makes sense if a logger running multiple times wants to register a single cleanup function.
        :return:
        """
        cache = self._get_teardown_cache()
        if cache.exists(name):
            if not silent:
                logger.debug("Post run fn with name '" + name + "' already exists. Skipped.")
        else:
            cache.add(name, post_fn)

    @cmd
    def register_teardown_fn(self, name, fn, error_message=None, nested=True, intermediate=True, order=0,
                             silent=True):
        """
        Register a new post_run_function by building it from given parameters.
        :param name: Name of the registration
        :param fn: Function to register
        :param error_message: Error message on failure.
        :param nested: Parameter if this function should be called on nested runs.
        :param intermediate: Parameter if this function should be called on a intermediate run.
        An intermediate run is a nested run managed specifically by pypads.
        :param order: Value defining the execution order for post run function.
        The lower the value the sooner a function gets executed.
        :param silent:
        :return:
        """
        self.register_teardown(name,
                               post_fn=RunTeardown(fn=fn, message=error_message, nested=nested,
                                                   intermediate=intermediate, order=order), silent=silent)

    @cmd
    def active_run(self):
        """
        Get the currently active run
        :return: Active run
        """
        return mlflow.active_run()

    @cmd
    def is_intermediate_run(self):
        """
        Check if the current run is an intermediate run.
        :return:
        """
        enclosing_run = self.pypads.cache.run_get("enclosing_run")
        return enclosing_run is not None

    @cmd
    def end_run(self):
        """
        End the current run and run its tearDown functions.
        :return:
        """
        run = self.active_run()

        consolidated_dict = self.pypads.cache.get('consolidated_dict', None)
        if consolidated_dict is not None:
            # Dump data to disk
            try_write_artifact("consolidated_log", consolidated_dict, write_format=WriteFormats.json)
            # self.log_mem_artifact("consolidated_log", consolidated_dict, write_format=WriteFormats.json)

        chached_fns = self._get_teardown_cache()
        fn_list = [v for i, v in chached_fns.items()]
        fn_list.sort(key=lambda t: t.order)
        for fn in fn_list:
            try:
                fn(self.pypads, _pypads_env=LoggerEnv(parameter=dict(), experiment_id=run.info.experiment_id,
                                               run_id=run.info.run_id))
            except (KeyboardInterrupt, Exception) as e:
                logger.warning("Failed running post run function " + fn.__name__ + " because of exception: " + str(e))

        mlflow.end_run()

        # --- Clean tmp files in disk cache after run ---
        folder = get_temp_folder(run)
        if os.path.exists(folder):
            import shutil
            shutil.rmtree(folder)
        # !-- Clean tmp files in disk cache after run ---

    # !--- run management ----
    @cmd
    def get_run(self, run_id=None):
        run_id = run_id or self.active_run().info.run_id
        return self.pypads.backend.mlf.get_run(run_id=run_id)

    @cmd
    def _get_metric_history(self, run):
        return {key: self.pypads.backend.mlf.get_metric_history(run.info.run_id, key) for key in run.data.metrics}

    @cmd
    def list_experiments(self, view_type: ViewType = ViewType.ALL):
        return self.pypads.backend.mlf.list_experiments(view_type)

    @cmd
    def list_run_infos(self, experiment_id, run_view_type: ViewType = ViewType.ALL):
        return self.pypads.backend.mlf.list_run_infos(experiment_id=experiment_id, run_view_type=run_view_type)

    @cmd
    def list_metrics(self, run_id=None):
        run = self.get_run(run_id)
        return self._get_metric_history(run)

    @cmd
    def list_parameters(self, run_id=None):
        run = self.get_run(run_id)
        return run.data.params

    @cmd
    def list_tags(self, run_id=None):
        run = self.get_run(run_id)
        return run.data.tags

    @cmd
    def list_artifacts(self, run_id=None, verbose=False):
        run = self.get_run(run_id)
        path = run.info.artifact_uri
        search = "Output"
        if verbose:
            search = ""
        return get_artifacts(path, search=search)

    @cmd
    def show_report(self, experiment_id=None):
        pass

    @cmd
    def list_logger_calls(self, run_id=None):
        run = self.get_run(run_id)
        path = run.info.artifact_uri
        return get_artifacts(path, search="Calls")

    @cmd
    def show_call_stack(self, run_id):
        pass

    @cmd
    def list_tracked_objects(self, run_id):
        run = self.get_run(run_id)
        path = run.info.artifact_uri
        return get_artifacts(path, search="TrackedObjects")

    @cmd
    def to_json(self, experiment_id):
        # Function to be called before ending the tracker
        from pypads.utils.logging_util import get_run_folder
        from pypads.utils.files_util import consolidate_run_output_files
        path = get_run_folder()
        consolidate_run_output_files(root_path=path)


class ApiPluginManager(ExtendableMixin):

    def __init__(self, *args, **kwargs):
        super().__init__(plugin_list=api_plugins)


pypads_api = PyPadsApi()


def api():
    """
    Returns classes of
    :return:
    """
    return inheritors(Cmd)



from collections.__init__ import OrderedDict

from pypads import logger
from pypads.app.call import CallAccessor, CallId, Call


# class CallMapping(CallAccessor):
#
#     def __init__(self, instance, _pypads_context, _pypads_wrappee, mapping):
#         super().__init__(instance, _pypads_context, _pypads_wrappee)
#         self._mapping = mapping
#
#     @property
#     def mapping(self):
#         return self._mapping


class CallTracker:
    """
    This class tracks the number of execution per instance of an object.
    """

    def __init__(self, pads):
        self._pads = pads
        self._call_stack = []

    def instance_call_number(self, accessor):
        function_calls: OrderedDict = self.function_call_dict(accessor.function_id)
        instance_id = accessor.instance_id

        call_items = list(function_calls.items())
        for instance_number in range(0, len(call_items)):
            # noinspection PyUnresolvedReferences
            key, value = call_items[instance_number]
            if key == instance_id:
                return instance_number
        return 0

    @property
    def call_stack(self):
        return self._call_stack

    def call_depth(self):
        return len(self._call_stack)

    def call_number(self, accessor: CallAccessor):
        """
        Get the number of calls of given call accessor
        :param accessor:
        :return:
        """
        return len(self.calls(accessor))

    def make_call_id(self, accessor: CallAccessor) -> CallId:
        """
        Returns the current call id. The id is built from: process_id, thread_id, defining_ctx_name, self_number/id,
        wrapped_fn_name, call_number
        :param accessor:
        :return:
        """
        return CallId(accessor.instance, accessor.context, accessor.wrappee, self.instance_call_number(accessor),
                      self.call_number(accessor))

    def current_call_number(self):
        """
        Get the current call number
        :return:
        """
        call = self._call_stack[-1]
        return self.call_number(call.accessor)

    def current_call(self):
        """
        Get the call_id of the current call
        :return:
        """
        return self._call_stack[-1] if len(self._call_stack) > 0 else None

    def current_process(self):
        return str(self._call_stack[-1].call_id.process) + "." + str(self._call_stack[-1].call_id.thread)

    def call_objects(self):
        """
        Get all call objects which are stored in our tracker.
        :return:
        """
        # Add call_objects if not exists in cache
        if not self._pads.cache.run_exists("call_objects"):
            self._pads.cache.run_add("call_objects", {})
        return self._pads.cache.run_get("call_objects")

    def function_call_dict(self, function_id) -> OrderedDict:
        call_objects = self.call_objects()
        if function_id not in call_objects:
            call_objects[function_id] = OrderedDict()
        return call_objects[function_id]

    def calls(self, accessor: CallAccessor):
        """
        Get all calls of given call accessor.
        :param accessor:
        :return:
        """
        instance_id = accessor.instance_id
        function_id = accessor.function_id

        function_calls = self.function_call_dict(function_id)
        if instance_id not in function_calls:
            function_calls[instance_id] = []

        return function_calls[instance_id]

    def has_call_identity(self, accessor: CallAccessor):
        for stored in self._call_stack:
            if stored.is_call_identity(accessor):
                return True
        return False

    def add(self, call: Call):
        """
        When a new call is triggered execute that function. This should be the first logging function called in each
        logging function stack.
        :return: A dict for holding information about the call.
        """
        self._call_stack.append(call)

        calls = self.calls(call.call_id)
        calls.append(call)
        return call

    def finish(self, call):
        if call in self._call_stack:
            call.finish()
            self._call_stack.remove(call)
            # TODO clear memory in call_objects?
        else:
            logger.error("Tried to finish call which is not on the stack. " + str(call))


def add_call(accessor):
    from pypads.app.pypads import get_current_pads
    pads = get_current_pads()
    call = Call(call_id=pads.call_tracker.make_call_id(accessor))
    return pads.call_tracker.add(call)


def finish_call(call):
    from pypads.app.pypads import get_current_pads
    pads = get_current_pads()
    return pads.call_tracker.finish(call)



from typing import List, Type

from pydantic import HttpUrl, BaseModel

from app.env import InjectionLoggerEnv
from pypads import logger
from pypads.app.injections.base_logger import TrackedObject, LoggerCall
from pypads.app.injections.injection import InjectionLogger
from pypads.model.models import TrackedObjectModel, ContextModel, OutputModel, \
    ParameterMetaModel
from pypads.utils.util import dict_merge


class ParametersTO(TrackedObject):
    """
    Tracking object class for model hyperparameters.
    """

    class HyperParameterModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/ModelHyperparameter"

        context: ContextModel = ...
        hyperparameters: List[ParameterMetaModel] = []

    def __init__(self, *args, tracked_by: LoggerCall, **kwargs):
        super().__init__(*args, tracked_by=tracked_by, **kwargs)
        self.context = tracked_by._logging_env.call.call_id.context

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.HyperParameterModel

    def _persist_parameter(self, key, value, description=None, type=""):
        name = self.context.reference + "." + key
        description = description or "Hyperparameter {} of context {}".format(name, self.context)
        meta = ParameterMetaModel(name=name,
                                  description=description,
                                  type=type)
        self.hyperparameters.append(meta)
        self._store_param(value, meta)


# def persist_parameter(_pypads_env, key, value):
#     try:
#         # TODO broken reference
#         try_mlflow_log(mlflow.log_param,
#                        _pypads_env.call.call_id.context.container.__name__ + "." + key + ".txt",
#                        value)
#     except Exception as e:
#         logger.warning(
#             "Couldn't track parameter. " + str(e) + " Trying to track with another name.")
#         try_mlflow_log(mlflow.log_param,
#                        str(_pypads_env.call) + "." + key + ".txt", value)


class ParametersILF(InjectionLogger):
    """
    Function logging the hyperparameters of the current pipeline object.
    """
    name = "ParametersLogger"
    uri = "https://www.padre-lab.eu/onto/hyperparameters-logger"

    class ParametersILFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/ParametersILF-Output"
        hyperparameters: ParametersTO.get_model_cls() = ...

        class Config:
            orm_mode = True

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.ParametersILFOutput

    def __post__(self, ctx, *args, _pypads_env: InjectionLoggerEnv, _logger_call, _logger_output, **kwargs):
        """
        Function logging the parameters of the current pipeline object function call.
        :param ctx:
        :param args:
        :param kwargs:
        :return:
        """
        hyper_params = ParametersTO(tracked_by=_logger_call)
        try:
            data = {}
            for mm in _pypads_env.mappings:
                if "data" in mm.mapping.values and 'hyper_parameters' in mm.mapping.values['data']:
                    data = dict_merge(data, mm.mapping.values['data']['hyper_parameters'])
            if len(data) > 0:
                for type, parameters in data.items():
                    for parameter in parameters:
                        key = parameter["name"]
                        if "path" in parameter and hasattr(ctx, parameter["path"]):
                            value = getattr(ctx, parameter["path"])
                            description = parameter.get('description', None)
                            type = parameter.get('kind_of_value', "")
                            hyper_params._persist_parameter(key, value, description, type)
                        else:
                            logger.warning("Couldn't access parameter " + key + " on " + str(ctx.__class__))
            else:
                logger.warning("No parameters are defined on the mapping file for " + str(
                    ctx.__class__) + ". Trying to extract by other means...")
                for key, value in ctx.get_params().items():
                    hyper_params._persist_parameter(key, value)
        except Exception as e:
            logger.error("Couldn't extract parameters on " + str(_pypads_env) + " due to " + str(e))
            # TODO what to with keras etc? Or define multiple loggers.
        finally:
            hyper_params.store(_logger_output, "hyperparameters")



import traceback
from abc import ABCMeta, abstractmethod
from typing import Type

from pydantic import BaseModel, HttpUrl

from app.env import InjectionLoggerEnv
from pypads import logger
from pypads.app.call import Call
from pypads.app.injections.base_logger import LoggerCall, Logger, LoggerExecutor, OriginalExecutor
from pypads.app.misc.mixins import OrderMixin, NoCallAllowedError
from pypads.model.models import InjectionLoggerCallModel, InjectionLoggerModel, MultiInjectionLoggerCallModel
from pypads.utils.util import inheritors


class InjectionLoggerCall(LoggerCall):

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return InjectionLoggerCallModel

    def __init__(self, *args, logging_env: InjectionLoggerEnv, **kwargs):
        super().__init__(*args, original_call=logging_env.call, logging_env=logging_env, **kwargs)


class InjectionLogger(Logger, OrderMixin, metaclass=ABCMeta):
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/injection-logger"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not hasattr(self, "_pre"):
            self._pre = LoggerExecutor(fn=self.__pre__)
        if not hasattr(self, "_post"):
            self._post = LoggerExecutor(fn=self.__post__)

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return InjectionLoggerModel

    def _base_path(self):
        return "InjectionLoggers/{}/".format(self.__class__.__name__)

    def __pre__(self, ctx, *args,
                _logger_call, _logger_output, _args, _kwargs, **kwargs):
        """
        The function to be called before executing the log anchor. the value returned will be passed on to the __post__
        function as **_pypads_pre_return**.


        :return: _pypads_pre_return
        """
        pass

    def __post__(self, ctx, *args, _logger_call, _pypads_pre_return, _pypads_result, _logger_output, _args, _kwargs,
                 **kwargs):
        """
        The function to be called after executing the log anchor.

        :param _pypads_pre_return: the value returned by __pre__.
        :param _pypads_result: the value returned by __call_wrapped__.

        :return: the wrapped function return value
        """
        pass

    def __real_call__(self, ctx, *args, _pypads_env: InjectionLoggerEnv, **kwargs):
        self.store_schema(self._base_path())

        _pypads_hook_params = _pypads_env.parameter

        logger_call = InjectionLoggerCall(logging_env=_pypads_env, created_by=self.store_schema())
        output = self.build_output()

        try:
            # Trigger pre run functions
            _pre_result, pre_time = self._pre(ctx, _pypads_env=_pypads_env,
                                              _logger_output=output,
                                              _logger_call=logger_call,
                                              _args=args,
                                              _kwargs=kwargs, **{**self.static_parameters, **_pypads_hook_params})
            logger_call.pre_time = pre_time

            # Trigger function itself
            _return, time = self.__call_wrapped__(ctx, _pypads_env=_pypads_env, _args=args, _kwargs=kwargs)
            logger_call.child_time = time

            # Trigger post run functions
            _post_result, post_time = self._post(ctx, _pypads_env=_pypads_env,
                                                 _logger_output=output,
                                                 _pypads_pre_return=_pre_result,
                                                 _pypads_result=_return,
                                                 _logger_call=logger_call,
                                                 _args=args,
                                                 _kwargs=kwargs, **{**self.static_parameters, **_pypads_hook_params})
            logger_call.post_time = post_time
        except Exception as e:
            logger_call.failed = str(e)
            output.set_failure_state(e)
            raise e
        finally:
            for fn in self.cleanup_fns(logger_call):
                fn(self, logger_call)
            logger_call.output = output.store(self._base_path())
            logger_call.store()
        return _return

    def __call_wrapped__(self, ctx, *args, _pypads_env: InjectionLoggerEnv, _args, _kwargs):
        """
        The real call of the wrapped function. Be carefull when you change this.
        Exceptions here will not be catched automatically and might break your workflow. The returned value will be passed on to __post__ function.

        :return: _pypads_result
        """
        _return, time = OriginalExecutor(fn=_pypads_env.callback)(*_args, **_kwargs)
        return _return, time

    def _handle_error(self, *args, ctx, _pypads_env, error, **kwargs):
        """
        Handle error for DefensiveCallableMixin
        :param args:
        :param ctx:
        :param _pypads_env:
        :param error:
        :param kwargs:
        :return:
        """
        try:
            raise error
        except NoCallAllowedError as e:

            # Call next wrapped callback if no call was allowed due to the settings or environment
            _pypads_hook_params = _pypads_env.parameter
            return self.__call_wrapped__(ctx, _pypads_env=_pypads_env, _args=args, _kwargs=kwargs,
                                         **_pypads_hook_params)
        except Exception as e:
            logger.error("Logging failed for " + str(self) + ": " + str(error) + "\nTrace:\n" + traceback.format_exc())

            # Try to call the original unwrapped function if something broke
            original = _pypads_env.call.call_id.context.original(_pypads_env.callback)
            if callable(original):
                try:
                    logger.error("Trying to recover from: " + str(e))
                    out = original(ctx, *args, **kwargs)
                    logger.success("Succeeded recovering on error : " + str(e))
                    return out
                except TypeError as e:
                    logger.error("Recovering failed due to: " + str(
                        e) + ". Trying to call without passed ctx. This might be due to an error in the wrapping.")
                    out = original(*args, **kwargs)
                    logger.success("Succeeded recovering on error : " + str(e))
                    return out
            else:

                # Original function was not accessiblete
                raise Exception("Couldn't fall back to original function for " + str(
                    _pypads_env.call.call_id.context.original_name(_pypads_env.callback)) + " on " + str(
                    _pypads_env.call.call_id.context) + ". Can't recover from " + str(error))


class MultiInjectionLoggerCall(LoggerCall):

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return MultiInjectionLoggerCallModel

    def __init__(self, *args, logging_env: InjectionLoggerEnv, **kwargs):
        super().__init__(*args, original_call=logging_env.call, logging_env=logging_env, **kwargs)
        self.call_stack = [logging_env.call]

    def add_call(self, call: Call):
        self.call_stack.append(call)


class MultiInjectionLogger(InjectionLogger):
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/multi-injection-logger"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # TODO teardown fn for output storage

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return MultiInjectionLoggerCallModel

    def _base_path(self):
        return "InjectionLoggers/{}/".format(self.__class__.__name__)

    def _get_call(self, logging_env: InjectionLoggerEnv):
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()
        if pads.cache.run_exists(id(self)):
            logger_call = pads.cache.run_get(id(self)).get('call')
            logger_call.add_call(logging_env.call)
            return logger_call
        else:
            self.store_schema(self._base_path())
            return MultiInjectionLoggerCall(logging_env=logging_env, created_by=self.store_schema())

    def _get_output(self, ):
        from pypads.app.pypads import get_current_pads
        pads = get_current_pads()
        if pads.cache.run_exists(id(self)):
            logger_output = pads.cache.run_get(id(self)).get('output')
            return logger_output
        else:
            return self.build_output()

    @staticmethod
    @abstractmethod
    def store(pads, *args, **kwargs):
        pass

    def __real_call__(self, ctx, *args, _pypads_env: InjectionLoggerEnv, **kwargs):
        _pypads_hook_params = _pypads_env.parameter

        logger_call = self._get_call(_pypads_env)
        output = self._get_output()

        try:
            # Trigger pre run functions
            _pre_result, pre_time = self._pre(ctx, _pypads_env=_pypads_env,
                                              _logger_output=output,
                                              _logger_call=logger_call,
                                              _args=args,
                                              _kwargs=kwargs, **{**self.static_parameters, **_pypads_hook_params})
            logger_call.pre_time += pre_time

            # Trigger function itself
            _return, time = self.__call_wrapped__(ctx, _pypads_env=_pypads_env, _args=args, _kwargs=kwargs)
            logger_call.child_time += time

            # Trigger post run functions
            _post_result, post_time = self._post(ctx, _pypads_env=_pypads_env,
                                                 _logger_output=output,
                                                 _pypads_pre_return=_pre_result,
                                                 _pypads_result=_return,
                                                 _logger_call=logger_call,
                                                 _args=args,
                                                 _kwargs=kwargs, **{**self.static_parameters, **_pypads_hook_params})
            logger_call.post_time += post_time
        except Exception as e:
            logger_call.failed = str(e)
            output.set_failure_state(e)
            raise e
        finally:
            for fn in self.cleanup_fns(logger_call):
                fn(self, logger_call)
            from pypads.app.pypads import get_current_pads
            pads = get_current_pads()
            pads.cache.run_add(id(self), {'call': logger_call, 'output': output, 'base_path': self._base_path()})
            pads.api.register_teardown_fn('{}_clean_up'.format(self.__class__.__name__), self.store)
        return _return


def logging_functions():
    """
    Find all injection functions defined in our imported context.
    :return:
    """
    return inheritors(InjectionLogger) + inheritors(MultiInjectionLogger)



from abc import ABCMeta
from typing import Type

from pydantic.main import BaseModel
from pydantic.networks import HttpUrl

# Default init_run fns
from pypads import logger
from pypads.app.injections.base_logger import LoggerCall, SimpleLogger
from pypads.app.misc.mixins import OrderMixin
from pypads.model.models import RunLoggerModel
from pypads.utils.util import inheritors


class RunLogger(SimpleLogger, OrderMixin, metaclass=ABCMeta):
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/run-logger"

    def build_call_object(self, _pypads_env, **kwargs):
        return LoggerCall(logging_env=_pypads_env, is_a="https://www.padre-lab.eu/onto/RunLoggerCall", **kwargs)

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return RunLoggerModel

    def _base_path(self):
        return "RunLoggers/"


class RunSetup(RunLogger, metaclass=ABCMeta):
    """
    This class should be used to define new pre run functions
    """
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/runsetup-logger"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __real_call__(self, *args, **kwargs):
        logger.debug("Called pre run function " + str(self))
        return super().__real_call__(*args, **kwargs)

    def _base_path(self):
        return super()._base_path() + "Setup/{}/".format(self.__name__)


class RunTeardown(RunLogger, metaclass=ABCMeta):
    """
    This class should be used to define new post run functions
    """
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/runteardown-logger"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __real_call__(self, *args, **kwargs):
        logger.debug("Called post run function " + str(self))
        return super().__real_call__(*args, **kwargs)

    def _base_path(self):
        return super()._base_path() + "Teardown/{}/".format(self.__name__)


def run_setup_functions():
    """
    Find all pre run functions defined in our imported context.
    :return:
    """
    return inheritors(RunSetup)


def run_teardown_functions():
    """
    Find all post run functions defined in our imported context.
    :return:
    """
    return inheritors(RunTeardown)



import glob
import os
from typing import List, Set, Tuple, Generator, Iterable

import yaml

from pypads import logger
from pypads.bindings.anchors import Anchor, get_anchor
from pypads.bindings.hooks import Hook
from pypads.importext.package_path import RegexMatcher, PackagePath, PackagePathMatcher, \
    SerializableMatcher, Package
from pypads.importext.versioning import LibSelector
from pypads.utils.util import find_package_version

default_mapping_file_paths = []
default_mapping_file_paths.extend(glob.glob(
    os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "bindings", "resources", "mapping", "**.yml"))))


class Mapping:
    """
    Mapping for an algorithm defined by a pypads mapping file
    """

    def __init__(self, matcher: PackagePathMatcher, in_collection, anchors, values):
        self._in_collection = in_collection
        self._values = values
        self._hooks = set()
        from pypads.bindings.hooks import Hook
        for a in anchors:
            self._hooks.add(Hook(a, self))
        self._matcher = matcher

    def is_applicable(self, ctx, obj):
        """
        Check if mapping fits next step on object of given context.
        :param ctx: Context of the object
        :param obj: Object to check
        :return:
        """
        if not hasattr(obj, "__name__"):
            return False
        reference = ctx.reference + "." + obj.__name__ if ctx is not None else obj.__name__
        return self._matcher.matches(PackagePath(reference))

    def applicable_filter(self, ctx):
        """
        Create a filter to check if the mapping is applicable
        :param ctx: Context on which we search an matching object
        :return:
        """

        def mapping_applicable_filter(name):
            if hasattr(ctx.container, name):
                try:
                    return self.is_applicable(ctx, getattr(ctx.container, name))
                except RecursionError as rerr:
                    logger.error("Recursion error on '" + str(
                        ctx) + "'. This might be because __get_attr__ is being wrapped. " + str(rerr))
            else:
                logger.debug("Can't access attribute '" + str(name) + "' on '" + str(ctx) + "'. Skipping.")
            return False

        return mapping_applicable_filter

    @property
    def in_collection(self):  # type: () -> MappingCollection
        return self._in_collection

    @in_collection.setter
    def in_collection(self, value):
        self._in_collection = value

    @property
    def reference(self):
        return self._matcher.serialize()

    @property
    def library(self) -> LibSelector:
        return self.in_collection.lib

    @property
    def hooks(self):  # type: () -> Set[Hook]
        return self._hooks

    @property
    def values(self):
        return self._values

    @property
    def matcher(self):
        return self._matcher

    def __str__(self):
        return "Mapping[" + str(self.reference) + ", lib=" + str(self.library) + ", hooks=" + str(self.hooks) + "]"

    def __eq__(self, other):
        return self.reference == other.reference and self.hooks == other.hooks and self.values == other.values

    def __hash__(self):
        return hash((self.reference, "|".join([str(h) for h in self.hooks]), str(self.values)))


class MappingCollection:
    def __init__(self, key, version, library):
        """
        Object holding a set of mappings related to a library
        :param key: Name / key of the collection
        :param version: Version of the collection
        :param library: Library information including library version constraint and name
        """
        self._mappings = {}
        self._name = key
        self._version = version
        self._lib = LibSelector.from_dict(library)

    @property
    def version(self):
        return self._version

    @property
    def lib(self):
        return self._lib

    @property
    def name(self):
        return self._name

    @property
    def mappings(self):
        return self._mappings

    def add_mapping(self, mapping: Mapping):
        """
        Add a mapping to the collection.
        :param mapping: Mapping to be added.
        :return:
        """
        path_map = self._mappings
        for segment in mapping.matcher.matchers:
            if segment not in path_map:
                path_map[segment] = dict()
            path_map = path_map[segment]
        if ":mapping" not in path_map:
            path_map[":mapping"] = []
        path_map[":mapping"].append(mapping)

    def _get_all_mappings(self, current_path=None):
        """
        Get all mappings stored behind place in the mapping dict.
        :param current_path: A reference to a sub dict in the mapping dict.
        :return: All mappings stored below the current path
        """
        mappings = []
        if current_path is None:
            current_path = self._mappings
        for k, v in current_path.items():
            if isinstance(k, str) and ":mapping" == k:
                mappings = mappings + v
            else:
                if isinstance(v, dict):
                    mappings = mappings + self._get_all_mappings(current_path=v)
        return mappings

    def find_mappings(self, segments, current_path=None):
        """
        Find all mappings matching given segments. The segments foo.bar for example are matched by foo.bar.a
        and foo.bar.{re:.*} etc.
        :param segments: Segments to look for
        :param current_path: Place in the mapping dict from which we are looking in mapping dict. We iterate through
        the map to find relevant mapping files.
        :return:
        """
        mappings = []
        if current_path is None:
            current_path = self._mappings
        if len(segments) > 0:
            if segments[0] in current_path:
                mappings = mappings + self.find_mappings(segments[1:], current_path[segments[0]])

            # For non PathSegments we can't use the hash lookup
            for s, v in current_path.items():
                if isinstance(s, RegexMatcher) and s.matches(segments[0]):
                    mappings = mappings + self.find_mappings(segments[1:], v)
        else:
            mappings = mappings + self._get_all_mappings(current_path)

        return mappings


def make_run_time_mapping_collection(lib):
    return MappingCollection("_pypads_runtime", "0.0.0", {"name": lib, "version": "0.0.0"})


class MappingSchema:
    """
    Schema holding the context of a mapping. This is used to build mappings iteratively.
    """

    def __init__(self, fragments, metadata, matcher: PackagePathMatcher, anchors, values):
        self._fragments = fragments
        self._metadata = metadata
        self._matcher = matcher
        if not isinstance(anchors, Iterable):
            anchors = [anchors]
        self._anchors = set()
        for anchor in anchors:
            self._anchors.add(get_anchor(anchor) or Anchor(anchor, "Runtime anchor. No description available."))
        self._values = values

    @property
    def fragments(self):
        return self._fragments

    @property
    def metadata(self):
        return self._metadata

    @property
    def matcher(self) -> PackagePathMatcher:
        return self._matcher

    @property
    def values(self):
        return self._values

    @property
    def anchors(self):
        return self._anchors

    def get(self, key):
        return self._values[key]

    def has(self, key):
        return key in self._values


class SerializedMapping(MappingCollection):
    """
    Mapping serialized by yaml.
    """

    def __init__(self, key, content):
        yml = yaml.load(content, Loader=yaml.SafeLoader)
        schema = MappingSchema(yml["fragments"] if "fragments" in yml else [], yml["metadata"], PackagePathMatcher(""),
                               set(), {})
        super().__init__(key, schema.metadata["version"], schema.metadata["library"])
        self._build_mappings(yml["mappings"], schema)

    def _build_mappings(self, node, schema: MappingSchema):
        """
        Build real mapping objects from a given yaml serialization.
        :param node: Yaml node holding the mappings
        :param schema: Current schema holding the context of the mapping. Initially this is an empty schema.
        :return: List of all mappings in the serialization
        """
        _fragments = []
        _children = []
        _anchors = set()
        _values = {}
        mappings = {}

        # replace fragments
        for k, v in node.items():
            if isinstance(k, SerializableMatcher):
                _children.append((k, v))
            elif k.startswith(":"):
                _children.append((PackagePathMatcher.unserialize(k[1:]), v))
            elif k.startswith(";"):
                _fragments.append(k[1:])
            elif k == "hooks":
                if isinstance(v, str):
                    _anchors.add(v)
                elif isinstance(v, List):
                    for anchor in v:
                        _anchors.add(anchor)
            elif k == "data":
                _values[k] = v

        for p in _fragments:
            for k, v in schema.fragments[p].items():
                node[k] = v
            del node[";" + p]

        schema = MappingSchema(schema.fragments, schema.metadata, matcher=schema.matcher,
                               anchors=_anchors.union(schema.anchors),
                               values={**schema.values, **_values})

        if len(_fragments) > 0:
            mappings = {**mappings, **self._build_mappings(node, schema)}
        else:
            if len(_children) > 0:
                for matcher, v in _children:
                    if schema.matcher:
                        matcher = schema.matcher + matcher
                    if not isinstance(matcher, PackagePathMatcher):
                        matcher = PackagePathMatcher(matcher)
                    mappings = {**mappings, **self._build_mappings(v, MappingSchema(schema.fragments, schema.metadata,
                                                                                    matcher=matcher,
                                                                                    anchors=schema.anchors,
                                                                                    values=schema.values))}
            elif schema.matcher.matchers is not None:
                self.add_mapping(
                    Mapping(schema.matcher, self, schema.anchors, schema.values))
        return mappings


class MappingFile(SerializedMapping):
    """
    Class referencing a file holding a mapping.
    """

    def __init__(self, path, name=None):
        with open(path) as f:
            if name is None:
                name = os.path.basename(f.name)
            data = f.read()
        super().__init__(name, data)


class MappingRegistry:
    """
    Class holding all the mappings
    """

    def __init__(self, pypads, *paths):
        # default paths
        self._pypads = pypads
        mapping_file_paths = []
        if pypads.config["include_default_mappings"]:
            # Use our with the package delivered mapping files
            mapping_file_paths.extend(glob.glob(os.path.join(pypads.folder, "bindings", "**.yml")))
            mapping_file_paths.extend(default_mapping_file_paths)
        if paths:
            mapping_file_paths.extend(paths)

        self._mappings = {}

        for path in mapping_file_paths:
            self.load_mapping(path)

    @staticmethod
    def from_params(pypads, mapping_input):
        """
        Method to create a mapping registry from input parameters given to the pypads app.
        :param pypads: Pypads app
        :param mapping_input: Input parameters of form: Path as string, Mapping object or a list of these
        :return: Mapping registry
        """
        paths = []
        mappings = []

        # Allow for referencing mappings via a path or via a direct reference to the object
        if not isinstance(mapping_input, List):
            mapping_input = [mapping_input]
        for m in mapping_input:
            if isinstance(m, str):
                paths.append(m)
            if isinstance(m, MappingCollection):
                mappings.append(m)

        # Pass paths to the constructor and add mappings to the registry manually
        registry = MappingRegistry(pypads, *paths)
        for mapping in mappings:
            if mapping:
                if isinstance(mapping, dict):
                    for key, mapping in mapping.items():
                        registry.add_mapping(mapping, key=key)
                else:
                    registry.add_mapping(mapping, key=mapping.lib)
        return registry

    def get_entries(self):  # type: () -> Generator[Tuple[LibSelector, MappingCollection]]
        """
        Method returning all entries in the mapping registry in a typed manner.
        :return: Generator of item tuples
        """
        for s, c in self._mappings.items():
            yield s, c

    def add_mapping(self, mapping: MappingCollection, key=None):
        """
        Method to add a mapping to the registry.
        :param mapping: Mapping collection to be added to the registry.
        :param key: Key for the mapping collection. TODO remove key?
        :return:
        """
        if key is None and isinstance(mapping, MappingFile):
            key = mapping.lib

        if key is None:
            logger.error(
                "Couldn't add mapping " + str(mapping) + " to the pypads mapping registry. Lib or key are undefined.")
        else:
            self._mappings[key] = mapping

    def load_mapping(self, path):
        """
        Load and add mapping at given path.
        :param path: Path to the mapping file.
        :return:
        """
        self.add_mapping(MappingFile(path))

    def get_libraries(self):
        """
        Find all supported libraries in the mapping registry.
        :return:
        """
        all_libs = set()
        for key, mapping in self._mappings.items():
            all_libs.add(mapping.lib)
        return all_libs

    def get_relevant_mappings(self, package: Package):
        """
        Function to find all relevant mappings. This produces a generator getting extended with found subclasses
        :return:
        """
        if any([package.path.segments[0] == s.name for s, _ in self.get_entries()]):
            lib_version = find_package_version(str(package.path.segments[0]))
            mappings = set()

            # Take only mappings which are fitting for versions if we have a selector
            if lib_version:
                lib_selector = LibSelector(name=str(package.path.segments[0]), constraint=lib_version)
                for k, collection in [(s, c) for s, c in self.get_entries() if s.allows_any(lib_selector)]:
                    for m in collection.find_mappings(package.path.segments):
                        mappings.add(m)

            # Otherwise just use all name fitting mappings
            else:
                for k, collection in [(s, c) for s, c in self.get_entries() if s.name == package.path.segments[0]]:
                    for m in collection.find_mappings(package.path.segments):
                        mappings.add(m)
            return mappings
        return set()


class MatchedMapping:
    """
    Class to denote a mapping hit. Given mapping matches given package path.
    """

    def __init__(self, mapping: Mapping, package_path: PackagePath):
        self._package_path = package_path
        self._mapping = mapping

    @property
    def package_path(self) -> PackagePath:
        return self._package_path

    @property
    def mapping(self) -> Mapping:
        return self._mapping

    def __hash__(self):
        return self._mapping.__hash__()

    def __str__(self):
        return "{}.{}".format(str(self._package_path), str(self._mapping))

    def __eq__(self, other):
        return self.mapping == other.mapping and self.package_path == other.package_path



import os
from typing import List, Type

from pydantic import BaseModel, HttpUrl

from pypads.app.injections.base_logger import LoggerCall, TrackedObject
from pypads.app.injections.injection import InjectionLogger
from pypads.model.models import TrackedObjectModel, OutputModel
from pypads.utils.logging_util import WriteFormats
from pypads.utils.util import local_uri_to_path, sizeof_fmt, PeriodicThread


def _get_cpu_usage():
    import psutil
    """
    cpu_usage = "CPU usage for cores:"
    for i, percentage in enumerate(psutil.cpu_percent(percpu=True)):
        cpu_usage += f"\nCore {i}: {percentage}%"
    cpu_usage += f"\nTotal CPU usage: {psutil.cpu_percent()}%"
    """
    cpu_usage = []
    for i, percentage in enumerate(psutil.cpu_percent(percpu=True)):
        cpu_usage.append(percentage)
    cpu_usage.append(psutil.cpu_percent())
    return cpu_usage


class CpuTO(TrackedObject):
    """
    Function logging the input parameters of the current pipeline object function call.
    """

    class CPUModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/CpuData"

        class CpuCoreModel(BaseModel):
            name: str = ...
            usage: List[float] = ...

            class Config:
                orm_mode = True

        content_format: WriteFormats = WriteFormats.text
        cpu_cores: List[CpuCoreModel] = []
        total_usage: List[str] = []
        period: float = ...

        class Config:
            orm_mode = True
            arbitrary_types_allowed = True

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.CPUModel

    def __init__(self, *args, tracked_by: LoggerCall, **kwargs):
        super().__init__(*args, tracked_by=tracked_by, **kwargs)

    def add_cpu_usage(self):
        cores = _get_cpu_usage()
        for idx, usage in enumerate(cores[:-1]):
            if idx >= len(self.cpu_cores):
                self.cpu_cores.append(self.CPUModel.CpuCoreModel(name='Core:' + str(idx), usage=[usage]))
            else:
                core = self.cpu_cores[idx]
                core.usage.append(usage)
        self.total_usage.append(cores[-1])

    def _get_artifact_path(self, name):
        return os.path.join(str(id(self)), "cpu_usage", name)


class CpuILF(InjectionLogger):
    """This logger extracts the cpu information of your machine."""
    name = "CPULogger"
    uri = "https://www.padre-lab.eu/onto/cpu-logger"

    class CpuILFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/CpuILF-Output"

        cpu_usage: CpuTO.get_model_cls() = ...

        class Config:
            orm_mode = True

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.CpuILFOutput

    _dependencies = {"psutil"}

    def __pre__(self, ctx, *args, _pypads_write_format=WriteFormats.text, _logger_call: LoggerCall, _logger_output,
                _pypads_period=1.0, _args, _kwargs,
                **kwargs):
        cpu_usage = CpuTO(tracked_by=_logger_call, content_format=_pypads_write_format)
        cpu_usage.period = _pypads_period

        def track_cpu_usage(to: CpuTO):
            to.add_cpu_usage()

        thread = PeriodicThread(target=track_cpu_usage, sleep=_pypads_period, args=(cpu_usage,))
        thread.start()

        # stop thread store disk_usage object
        def cleanup_thread(logger, _logger_call):
            thread.join()
            cpu_usage.store(_logger_output, key="cpu_usage")

        self.register_cleanup_fn(_logger_call, fn=cleanup_thread)


class RamTO(TrackedObject):
    """
    Function logging the memory information of the current pipeline object function call.
    """

    class RAMModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/RamData"

        class MemoryModel(BaseModel):
            used: List[int] = ...
            free: List[int] = ...
            percentage: List[float] = ...
            type: str = ...

        content_format: WriteFormats = WriteFormats.json
        virtual_memory: MemoryModel = None
        swap_memory: MemoryModel = None
        period: float = ...

        class Config:
            orm_mode = True
            arbitrary_types_allowed = True

    def __init__(self, *args, tracked_by: LoggerCall, **kwargs):
        super().__init__(*args, tracked_by=tracked_by, **kwargs)

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.RAMModel

    def to_string(self):
        memory_usage = "Memory usage:"
        for item in [self.virtual_memory, self.swap_memory]:
            memory_usage += f"\n\tType:{item.type}"
            memory_usage += f"\n\t\tUsed:{sizeof_fmt(item.used)}"
            memory_usage += f"\n\t\tUsed:{sizeof_fmt(item.free)}"
            memory_usage += f"\n\t\tPercentage:{item.percentage}%"
        return memory_usage

    def add_memory_usage(self):
        ram_info, swap_info = _get_memory_usage()

        used = ram_info.get('used')
        free = ram_info.get('free')
        percent = ram_info.get('percent')
        if self.virtual_memory is None:
            self.virtual_memory = self.RAMModel.MemoryModel(name='ram_usage', used=[used], free=[free],
                                                            percentage=[percent], type="RAM")
        else:
            self.virtual_memory.used.append(used)
            self.virtual_memory.free.append(free)
            self.virtual_memory.percentage.append(percent)

        used = swap_info.get('used')
        free = swap_info.get('free')
        percent = swap_info.get('percent')
        if self.swap_memory is None:
            self.swap_memory = self.RAMModel.MemoryModel(name='swap_usage', used=[used], free=[free],
                                                         percentage=[percent], type="Swap")
        else:
            self.swap_memory.used.append(used)
            self.swap_memory.free.append(free)
            self.swap_memory.percentage.append(percent)

    def _get_artifact_path(self, name):
        return os.path.join(str(id(self)), "memory_usage", name)


class RamILF(InjectionLogger):
    """
    This function only writes an information of a constructor execution to the stdout.
    """

    name = "RAMLogger"
    uri = "https://www.padre-lab.eu/onto/ram-logger"

    class RamILFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/RamILF-Output"

        memory_usage: RamTO.RAMModel = ...

        class Config:
            orm_mode = True

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.RamILFOutput

    _dependencies = {"psutil"}

    def __pre__(self, ctx, *args, _pypads_write_format=WriteFormats.json, _logger_call: LoggerCall, _logger_output,
                _pypads_period=1.0, _args, _kwargs, **kwargs):
        memory_usage = RamTO(tracked_by=_logger_call, content_format=_pypads_write_format)

        def track_mem_usage(to: RamTO):
            to.add_memory_usage()

        memory_usage.period = _pypads_period
        thread = PeriodicThread(target=track_mem_usage, args=(memory_usage,))
        thread.start()

        # stop thread store disk_usage object
        def cleanup_thread(logger, _logger_call):
            thread.join()
            memory_usage.store(_logger_output, key="memory_usage")

        self.register_cleanup_fn(_logger_call, fn=cleanup_thread)


def _get_memory_usage():
    import psutil
    memory = psutil.virtual_memory()
    swap = psutil.swap_memory()
    return dict(memory._asdict()), dict(swap._asdict())


class DiskTO(TrackedObject):
    """
    Function logging the input parameters of the current pipeline object function call.
    """

    class DiskModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/DiskData"

        class PartitionModel(BaseModel):
            name: str = ...
            device: str = ...
            file_system: str = ...
            mount_point: str = ...
            free: List[int] = ...
            used: List[int] = ...
            percentage: List[float] = ...

            class Config:
                orm_mode = True

        content_format: WriteFormats = WriteFormats.text
        partitions: List[PartitionModel] = []

        period: float = ...
        path: str = ...

        class Config:
            orm_mode = True
            arbitrary_types_allowed = True

    def __init__(self, *args, tracked_by: LoggerCall, path=None, **kwargs):
        self.path = path
        super().__init__(*args, tracked_by=tracked_by, **kwargs)

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.DiskModel

    def init_disk_usage(self):
        value = _get_disk_usage(self.path)
        for partition, info in value.get('partitions', dict()).items():
            _name = partition
            file_system = info.get('FileSystem')
            device = info.get('device')
            mount_point = info.get('MountPoint')
            free = [info.get('free')]
            used = [info.get('used')]
            percent = [info.get('percentage')]
            self.partitions.append(self.DiskModel.PartitionModel(name=_name, file_system=file_system,
                                                                 device=device, mount_point=mount_point, free=free,
                                                                 used=used,
                                                                 percentage=percent))

    def add_disk_usage(self):
        value = _get_disk_usage(self.path)
        for partition, info in value.get('partitions', dict()).items():
            for pm in self.partitions:
                if pm.name == partition:
                    pm.free.append(info.get('free'))
                    pm.used.append(info.get('used'))
                    pm.percentage.append(info.get('percentage'))

    def _get_artifact_path(self, name):
        return os.path.join(str(id(self)), "disk_usage", name)


class DiskILF(InjectionLogger):
    """
    This function only writes an information of a constructor execution to the stdout.
    """

    name = "DiskLogger"
    uri = "https://www.padre-lab.eu/onto/disk-logger"

    class DiskILFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/DiskILF-Output"
        disk_usage: List[DiskTO.DiskModel] = []

        class Config:
            orm_mode = True

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.DiskILFOutput

    @classmethod
    def _needed_packages(cls):
        return ["psutil"]

    def __pre__(self, ctx, *args, _pypads_write_format=WriteFormats.text, _logger_call: LoggerCall, _logger_output,
                _pypads_disk_usage=None, _pypads_period=1.0, _args, _kwargs,
                **kwargs):
        from pypads.app.base import PyPads
        from pypads.app.pypads import get_current_pads
        pads: PyPads = get_current_pads()
        if _pypads_disk_usage is None:
            _pypads_disk_usage = [local_uri_to_path(pads.uri)]

        def track_disk_usage(to: DiskTO):
            to.add_disk_usage()

        for p in _pypads_disk_usage:
            disk_usage_to = DiskTO(tracked_by=_logger_call, content_format=_pypads_write_format, path=p)
            disk_usage_to.init_disk_usage()
            disk_usage_to.period = _pypads_period
            thread = PeriodicThread(target=track_disk_usage, sleep=_pypads_period, args=(disk_usage_to,))
            thread.start()

            # stop thread store disk_usage object
            def cleanup_thread(logger, _logger_call):
                thread.join()
                disk_usage_to.store(_logger_output, key="disk_usage")

            self.register_cleanup_fn(_logger_call, fn=cleanup_thread)


def _get_disk_usage(path):
    import psutil
    # See https://www.thepythoncode.com/article/get-hardware-system-information-python
    disk_usage = psutil.disk_usage(path)
    output_dict = dict()
    output_dict['free'] = disk_usage.free
    output_dict['used'] = disk_usage.used
    output_dict['percentage'] = disk_usage.percent

    partition_dict = dict()
    partitions = psutil.disk_partitions()
    for idx, partition in enumerate(partitions):
        temp_dict = dict()
        temp_dict['device'] = partition.device
        temp_dict['MountPoint'] = partition.mountpoint
        temp_dict['FileSystem'] = partition.fstype
        partition_usage = None
        try:
            partition_usage = psutil.disk_usage(partition.mountpoint)
        except PermissionError:
            # output_ += f"\n\t\t Busy!"
            continue

        temp_dict['free'] = partition_usage.free
        temp_dict['used'] = partition_usage.used
        temp_dict['percentage'] = partition_usage.percent

        partition_dict['partition' + str(idx)] = temp_dict

    output_dict['partitions'] = partition_dict
    return output_dict



import os

from pypads import logger
from pypads.utils.util import is_package_available


def check_determinism():
    from pypads.app.pypads import get_current_pads
    pads = get_current_pads()
    if is_package_available('tensorflow'):
        import tensorflow
        tf_version = tensorflow.version.VERSION
        if tensorflow.match("(1\.(14|15)|2\.0)", tf_version):
            if "TF_USE_CUDNN_AUTOTUNE" in os.environ:
                logger.warning(
                    "When using TF auto-tuning of cuDNN convolution algorithms your experiment might"
                    " be non-deterministic.")
                pads.api.set_tag("non-determinism", "CUDNN_AUTOTUNE")

            if ("TF_CUDNN_DETERMINISTIC" not in os.environ or (not os.environ["TF_CUDNN_DETERMINISTIC"] and os.environ[
                "TF_CUDNN_DETERMINISTIC"] is not 1)):
                if not is_package_available("tfdeterminism"):
                    logger.warning(
                        "Your experiment might include a gpu-specific sources of non-determinism."
                        " See https://github.com/NVIDIA/tensorflow-determinism")
                    pads.api.set_tag("non-determinism",
                                     "TF auto-tuning of cuDNN convolution algorithms (see multi-algo note)")

    # TODO Check for different settings (see ttps://github.com/NVIDIA/tensorflow-determinism)

# def _check_autotuning(self):
#
# def _check_backprop_weight_gradients(self):
#
# def _check_backprop_data_gradients(self):
#
# def _check_cuDNN_maxpooling_backprop(self):
#
# def _check_cuDNN_CTC_loss(self):
#
# def _check_bias_add_backprop(self):
#
# def _check_XLA_reductions_on_GPU(self):
#
# def _check_fused_softmax_cross_entropy_ops_backprop(self):
#     logger.warning("When using TF auto-tuning of cuDNN convolution algorithms your experiment might
#     be non-deterministic.")



import os
import time
import uuid
from typing import List, Optional

from pydantic import BaseModel, Field, HttpUrl, root_validator

from pypads.utils.logging_util import WriteFormats
from pypads.utils.util import get_experiment_id, get_run_id


class OntologyEntry(BaseModel):
    """
    Object representing an (potential) entry in a knowledge base
    """
    uri: HttpUrl = ...


class LibraryModel(BaseModel):
    """
    Representation of a package or library
    """
    name: str = ...
    version: str = ...
    extracted: bool = False


class LibSelectorModel(BaseModel):
    """
    Representation of a selector for a package
    """
    name: str = ...  # Name of the package. Either a direct string or a regex.
    constraint: str  # Constraint for the version number
    regex: bool = False  # Flag if the name of the selector is to be considered as a regex
    specificity: int  # How specific the selector is ( important for a css like mapping of multiple selectors)

    def __hash__(self):
        return hash((self.name, self.constraint, self.specificity))

    class Config:
        orm_mode = True


class RunObjectModel(OntologyEntry):
    """
    Base object for tracked objects that manage metadata. A MetadataEntity manages and id and a dict of metadata.
    The metadata should contain all necessary non-binary data to describe an entity.
    """
    experiment_id: Optional[str] = Field(default_factory=get_experiment_id)
    run_id: Optional[str] = Field(default_factory=get_run_id)
    uid: uuid.UUID = Field(default_factory=uuid.uuid4)
    created_at: float = Field(default_factory=time.time)
    is_a: HttpUrl = ...
    uri: HttpUrl = None

    @root_validator
    def set_default_uri(cls, values):
        if values['uri'] is None:
            values['uri'] = f"{values['is_a']}#{values['uid']}"
        return values

    def store(self):
        """
        Function to store the object as json into an artifact
        :return:
        """
        from pypads.app.pypads import get_current_pads
        from pypads.utils.logging_util import WriteFormats
        get_current_pads().api.log_mem_artifact("{}#{}".format(self.__class__.__name__, self.uid), self.json(),
                                                WriteFormats.json.value,
                                                path=os.path.join(self.__class__.__name__, str(self.uid)))


class LoggerModel(RunObjectModel):
    """
    A reference object for a logger.
    """
    name: str = "GenericTrackingFunction"
    uid: uuid.UUID = Field(default_factory=uuid.uuid4)
    uri: HttpUrl = "https://www.padre-lab.eu/onto/generic-tracking-function"
    dependencies: List[LibSelectorModel] = {}
    supported_libraries: List[LibSelectorModel] = ...
    allow_nested: bool = True
    allow_intermediate: bool = True

    class Config:
        orm_mode = True


class RunLoggerModel(LoggerModel):
    """
    Tracking function being executed on run teardown or setup
    """
    name: str = "GenericRunLogger"
    uri: HttpUrl = "https://www.padre-lab.eu/onto/generic-run-logger"

    class Config:
        orm_mode = True


class InjectionLoggerModel(LoggerModel):
    """
    Tracking function being exectured on injection hook execution
    """
    name: str = "GenericInjectionLogger"
    uri: HttpUrl = "https://www.padre-lab.eu/onto/generic-injection-logger"

    class Config:
        orm_mode = True


class ContextModel(BaseModel):
    reference: str = ...  # Path to the context e.g.: sklearn.tree.tree.DecisionTree

    class Config:
        orm_mode = True


class FunctionReferenceModel(BaseModel):
    fn_name: str = ...  # Function name on the given context e.g.: sklearn.tree.tree.DecisionTree.fit
    context: ContextModel = ...  # Context on which function was defined

    class Config:
        orm_mode = True


class CallAccessorModel(FunctionReferenceModel):
    instance_id: int = ...  # Instance id of the instance on which a call was done

    class Config:
        orm_mode = True


class CallIdModel(CallAccessorModel):
    process: int = ...  # Process of the call
    thread: int = ...  # Thread of the call
    instance_number: int = ...  # Number of the call on instance
    call_number: int = ...  # Plain number of the call

    class Config:
        orm_mode = True


class CallModel(RunObjectModel):
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/Call"
    call_id: CallIdModel = ...  # Id of the call
    finished: bool = False

    class Config:
        orm_mode = True


class MetadataModel(BaseModel):
    path: str = ...
    description: str = ...
    format: WriteFormats = ...


class MetricMetaModel(BaseModel):
    name: str = ...
    description: str = ...
    step: int = ...


class ParameterMetaModel(BaseModel):
    name: str = ...
    description: str = ...
    type: str = ...


class ArtifactMetaModel(BaseModel):
    path: str = ...
    description: str = ...
    format: WriteFormats = ...


class TagMetaModel(BaseModel):
    name: str = ...
    description: str = ...


class LoggerCallModel(RunObjectModel):
    """
    Holds meta data about a logger execution
    """
    failed: Optional[str] = None
    created_by: str = ...  # path to json of LoggerModel
    execution_time: Optional[float] = ...
    output: Optional[str] = ...  # path to json of the OutputModel of the logger
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/LoggerCall"

    class Config:
        orm_mode = True


class InjectionLoggerCallModel(LoggerCallModel):
    """
    Holds meta data about an injection logger execution
    """
    pre_time: Optional[float] = ...
    post_time: Optional[float] = ...
    child_time: Optional[float] = ...
    original_call: CallModel = ...  # Triggered by following call
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/InjectionLoggerCall"
    execution_time: Optional[float] = None

    @root_validator
    def set_default_execution_time(cls, values):
        if values['execution_time'] is None:
            if values['pre_time'] is not None and values['post_time'] is not None:
                values['execution_time'] = values['pre_time'] + values['post_time']
        return values

    class Config:
        orm_mode = True


class MultiInjectionLoggerCallModel(InjectionLoggerCallModel):
    """Holds meta data about an injection logger multiple execution"""
    call_stack: List[CallModel] = ...
    pre_time: Optional[float] = 0.0
    post_time: Optional[float] = 0.0
    child_time: Optional[float] = 0.0
    is_a : HttpUrl = "https://www.padre-lab.eu/onto/MultiInjectionLoggerCall"

    class Config:
        orm_mode = True


class OutputModel(RunObjectModel):
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/LoggerOutput"

    failed: Optional[str] = None

    class Config:
        orm_mode = True


class EmptyOutput(OutputModel):
    is_a: HttpUrl = "https://www.padre-lab.eu/onto/EmptyLoggerOutput"

    class Config:
        orm_mode = True


class TrackedObjectModel(RunObjectModel):
    """
    Data of a tracking object.
    """
    tracked_by: LoggerCallModel = ...

    class Config:
        orm_mode = True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypads',
 'pypads.app',
 'pypads.app.backends',
 'pypads.app.injections',
 'pypads.app.misc',
 'pypads.bindings',
 'pypads.bindings.resources',
 'pypads.bindings.resources.mapping',
 'pypads.importext',
 'pypads.importext.semver',
 'pypads.importext.wrapping',
 'pypads.injections',
 'pypads.injections.analysis',
 'pypads.injections.loggers',
 'pypads.injections.loggers.mlflow',
 'pypads.injections.setup',
 'pypads.model',
 'pypads.parallel',
 'pypads.utils']

package_data = \
{'': ['*']}

install_requires = \
['boltons>=19.3.0,<20.0.0',
 'cloudpickle>=1.3.0,<2.0.0',
 'loguru>=0.4.1,<0.5.0',
 'mlflow>=1.6.0,<2.0.0',
 'pydantic>=1.5.1,<2.0.0']

setup_kwargs = {
    'name': 'pypads',
    'version': '0.3.1',
    'description': 'PyPaDS aims to to add tracking functionality to machine learning libraries.',
    'long_description': '# PyPads\nBuilding on the [MLFlow](https://github.com/mlflow/mlflow/) toolset this project aims to extend the functionality for MLFlow, increase the automation and therefore reduce the workload for the user. The production of structured results is an additional goal of the extension.\n\n[![Documentation Status](https://readthedocs.org/projects/pypads/badge/?version=latest)](https://pypads.readthedocs.io/en/latest/?badge=latest)\n[![PyPI version](https://badge.fury.io/py/pypads.svg)](https://badge.fury.io/py/pypads)  \n\n<!--- ![Build status](https://gitlab.padim.fim.uni-passau.de/RP-17-PaDReP/ontopads/badges/master/pipeline.svg) --->\n\n# Intalling\nThis tool requires those libraries to work:\n\n    Python (>= 3.6),\n    cloudpickle (>= 1.3.3),\n    mlflow (>= 1.6.0),\n    boltons (>= 19.3.0),\n    loguru (>=0.4.1)\n    \nPyPads only support python 3.6 and higher. To install pypads run this in you terminal\n\n**Using source code**\n\nFirst, you have to install **poetry** \n\n    pip install poetry\n    poetry build (in the root folder of the repository pypads/)\n\nThis would create two files under pypads/dist that can be used to install,\n\n    pip install dist/pypads-X.X.X.tar.gz\n    OR\n    pip install dist/pypads-X.X.X-py3-none-any.whl\n    \n \n**Using pip ([PyPi release](https://pypi.org/project/pypads/))**\n\nThe package can be found on PyPi in following [project](https://pypi.org/project/pypads/).\n\n    pip install pypads\n\n### Tests\nThe unit tests can be found under \'test/\' and can be executed using\n\n    poetry run pytest test/\n\n# Documentation\n\nFor more information, look into the [official documentation of PyPads](https://pypads.readthedocs.io/en/latest/).\n\n# Getting Started\n         \n### Usage example\npypads is easy to use. Just define what is needed to be tracked in the config and call PyPads.\n\nA simple example looks like the following,\n```python\nfrom pypads.app.base import PyPads\n# define the configuration, in this case we want to track the parameters, \n# outputs and the inputs of each called function included in the hooks (pypads_fit, pypads_predict)\nhook_mappings = {\n    "parameters": {"on": ["pypads_fit"]},\n    "output": {"on": ["pypads_fit", "pypads_predict"]},\n    "input": {"on": ["pypads_fit"]}\n}\n# A simple initialization of the class will activate the tracking\nPyPads(hooks=hook_mappings)\n\n# An example\nfrom sklearn import datasets, metrics\nfrom sklearn.tree import DecisionTreeClassifier\n\n# load the iris datasets\ndataset = datasets.load_iris()\n\n# fit a model to the data\nmodel = DecisionTreeClassifier()\nmodel.fit(dataset.data, dataset.target) # pypads will track the parameters, output, and input of the model fit function.\n# get the predictions\npredicted = model.predict(dataset.data) # pypads will track only the output of the model predict function.\n```\n        \n        \nThe used hooks for each event are defined in the mapping file where each hook represents the functions to listen to.\nUsers can use regex for goruping functions and even provide paths to hook functions.\nIn the [sklearn mapping](pypads/bindings/resources/mapping/sklearn_0_19_1.yml) YAML file, an example entry would be:\n```yaml\nfragments:\n  default_model:\n    !!python/pPath __init__:\n      hooks: "pypads_init"\n    !!python/rSeg (fit|.fit_predict|fit_transform)$:\n      hooks: "pypads_fit"\n    !!python/rSeg (fit_predict|predict|score)$:\n      hooks: "pypads_predict"\n    !!python/rSeg (fit_transform|transform)$:\n      hooks: "pypads_transform"\n\nmappings:\n  !!python/pPath sklearn:\n    !!python/pPath base.BaseEstimator:\n      ;default_model: ~\n```\nFor instance, "pypads_fit" is an event listener on any fit, fit_predict and fit_transform call made by the tracked model class which is in this case **BaseEstimator** that most estimators inherits from.\n\nUsing no custom yaml types and no fragments the mapping file would be equal to following definition:\n```yaml\nmappings:\n  :sklearn:\n    :base.BaseEstimator:\n        :__init__:\n          hooks: "pypads_init"\n        :{re:(fit|.fit_predict|fit_transform)$}:\n          hooks: "pypads_fit"\n        :{re:(fit_predict|predict|score)$}:\n          hooks: "pypads_predict"\n        :{re:(fit_transform|transform)$}:\n          hooks: "pypads_transform"\n```\n\n# Acknowledgement\nThis work has been partially funded by the **Bavarian Ministry of Economic Affairs, Regional Development and Energy** by means of the funding programm **"Internetkompetenzzentrum Ostbayern"** as well as by the **German Federal Ministry of Education and Research** in the project **"Provenance Analytics"** with grant agreement number *03PSIPT5C*.\n',
    'author': 'Thomas Weißgerber',
    'author_email': 'thomas.weissgerber@uni-passau.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.padre-lab.eu/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.0,<4.0.0',
}


setup(**setup_kwargs)



import inspect
import sys
import types
from functools import wraps
from importlib._bootstrap_external import PathFinder
# noinspection PyUnresolvedReferences
from multiprocessing import Value

from pypads import logger
from pypads.importext.mappings import Mapping, MatchedMapping
from pypads.importext.package_path import PackagePath, PackagePathMatcher, Package
from pypads.importext.wrapping.base_wrapper import Context


def _add_found_class(mapping):
    from pypads.app.pypads import get_current_pads
    return get_current_pads().mapping_registry.add_found_class(mapping)


def _get_relevant_mappings(package: Package):
    from pypads.app.pypads import get_current_pads
    return get_current_pads().mapping_registry.get_relevant_mappings(package)


def _add_inherited_mapping(clazz, super_class):
    from pypads.app.pypads import get_current_pads
    found_mappings = set()
    if clazz.__name__ not in get_current_pads().wrap_manager.class_wrapper.punched_class_names:
        if hasattr(super_class, "_pypads_mapping_" + super_class.__name__):
            for matched_mapping in getattr(super_class, "_pypads_mapping_" + super_class.__name__):
                """
                Build the package path matcher by looking at the superclass matched_mappings and deconstructing the 
                package_path of it. Taking from the matcher unmatched parts and adding them to the module and qualname.
                """
                found_mapping = Mapping(PackagePathMatcher(
                    ".".join(filter(lambda s: len(s) > 0,
                                    [clazz.__module__, clazz.__qualname__,
                                     ".".join([h.serialize() for h in
                                               matched_mapping.mapping.matcher.matchers[
                                               len(matched_mapping.package_path.segments):]])]))),
                    matched_mapping.mapping.in_collection, {h.anchor for h in matched_mapping.mapping.hooks},
                    matched_mapping.mapping.values)
                found_mappings.add(found_mapping)
    return found_mappings


def duck_punch_loader(spec):
    if not spec.loader:
        return spec

    original_exec = spec.loader.exec_module

    @wraps(original_exec)
    def exec_module(self, module, execute=original_exec):
        out = execute(module)

        # History to check if a class inherits a wrapping intra-module
        mro_entry_history = {}

        from pypads.app.pypads import current_pads
        reference = module.__name__

        if current_pads:
            # TODO we might want to make this configurable/improve performance.
            #  This looks at every imported class and every mapping.
            # On execution of a module we search for relevant mappings
            # For every var on module
            try:
                members = inspect.getmembers(module,
                                             lambda x: hasattr(x, "__module__") and x.__module__ == module.__name__)
            except Exception as e:
                logger.debug(
                    "getmembers of inspect failed on module '" + str(module.__name__) + "' with expection" + str(
                        e) + ". Falling back to dir to get the members of the module.")
                members = [(name, getattr(module, name)) for name in dir(module)]

            for name, obj in members:
                if obj is not None:
                    obj_ref = ".".join([reference, name])
                    package = Package(module, PackagePath(obj_ref))

                    # Skip modules if they are from another package for now
                    if inspect.ismodule(obj):
                        if not module.__name__.split(".")[0] == obj.__name__.split(".")[0]:
                            continue

                    mappings = set()
                    if inspect.isclass(obj) and hasattr(obj, "mro"):
                        try:

                            # Look at the MRO and add classes to be punched which inherit from our punched classes
                            mro_ = obj.mro()[1:]
                            for entry in mro_:
                                if entry not in mro_entry_history.keys():
                                    mro_entry_history[entry] = [obj]
                                else:
                                    mro_entry_history[entry].append(obj)
                                if hasattr(entry, "_pypads_mapping_" + entry.__name__):
                                    found_mappings = _add_inherited_mapping(obj, entry)
                                    mappings = mappings.union(found_mappings)
                        except Exception as e:
                            logger.debug("Skipping some superclasses of " + str(obj) + ". " + str(e))
                    mappings = mappings.union(_get_relevant_mappings(package))
                    if len(mappings) > 0:
                        current_pads.wrap_manager.wrap(obj, Context(module, reference),
                                                       {MatchedMapping(mapping, package.path) for mapping in mappings})
        return out

    spec.loader.exec_module = types.MethodType(exec_module, spec.loader)
    return spec


class PyPadsFinder(PathFinder):
    """
    Import lib extension. This finder provides a special loader if mapping files contain the object.
    """

    def find_spec(cls, fullname, path=None, target=None):
        if fullname not in sys.modules:
            path_ = sys.meta_path[
                    [i for i in range(len(sys.meta_path)) if isinstance(sys.meta_path[i], PyPadsFinder)].pop() + 1:]
            i = iter(path_)
            spec = None
            try:
                # Find a valid importer for the given fullname on the meta_path
                importer = None
                while not spec:
                    importer = next(i)
                    if hasattr(importer, "find_spec"):
                        spec = importer.find_spec(fullname, path)

                # Use the importer as a real importer but wrap it in the PyPadsLoader
                if spec and importer and spec.loader:
                    return duck_punch_loader(spec)
            except StopIteration:
                pass


def extend_import_module():
    """
    Function to add the custom import logic to the python importlib execution
    :return:
    """
    path_finder = [i for i in range(len(sys.meta_path)) if
                   "_frozen_importlib_external.PathFinder" in str(sys.meta_path[i])]
    sys.meta_path.insert(path_finder.pop(), PyPadsFinder())



import os
from typing import Union, Type, Optional

from pydantic import HttpUrl, BaseModel

from pypads import logger
from pypads.app.injections.base_logger import TrackedObject
from pypads.app.injections.injection import InjectionLogger
from pypads.model.models import TrackedObjectModel, ArtifactMetaModel, MetricMetaModel, OutputModel
from pypads.utils.logging_util import WriteFormats


class MetricTO(TrackedObject):
    """
    Tracking object for metrics computed in the workflow.
    """

    class MetricModel(TrackedObjectModel):
        uri: HttpUrl = "https://www.padre-lab.eu/onto/Metric"

        name: str = ...  # Metric name
        to_artifact: bool = False
        value: Union[MetricMetaModel, ArtifactMetaModel] = ...

        class Config:
            orm_mode = True

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return cls.MetricModel

    def store_value(self, value, step):
        self.name = self.tracked_by.original_call.call_id.context.container.__name__ + "." + \
                    self.tracked_by.original_call.call_id.wrappee.__name__

        if isinstance(value, float):
            self.value = MetricMetaModel(name=self.name, step=step,
                                         description="The metric returned by {}".format(self.name))
            self._store_metric(value, self.value)
            return True
        else:
            logger.warning("Mlflow metrics have to be doubles. Could log the return value of type '" + str(
                type(
                    value)) + "' of '" + self.name + "' as artifact instead.")
            if self.to_artifact:
                path = os.path.join(self._base_path(), self._get_artifact_path(self.name))
                self.value = ArtifactMetaModel(path=path, description="", format=WriteFormats.text)
                self._store_artifact(value, self.value)
                return True
        return False


class MetricILF(InjectionLogger):
    """
    Function logging the wrapped metric function
    """
    name = "Metric Injection Logger"
    uri = "https://www.padre-lab.eu/onto/metric-logger"

    class MetricILFOutput(OutputModel):
        is_a: HttpUrl = "https://www.padre-lab.eu/onto/MetricILF-Output"
        metric: Optional[MetricTO.MetricModel] = None

        class Config:
            orm_mode = True

    @classmethod
    def output_schema_class(cls) -> Type[OutputModel]:
        return cls.MetricILFOutput

    def __post__(self, ctx, *args, _pypads_artifact_fallback=False, _logger_call, _logger_output, _pypads_result, **kwargs):
        """

        :param ctx:
        :param args:
        :param _pypads_artifact_fallback: Write to artifact if metric can not be logged as an double value into mlflow
        :param _pypads_result:
        :param kwargs:
        :return:
        """

        result = _pypads_result
        metric = MetricTO(tracked_by=_logger_call,
                          to_artifact=_pypads_artifact_fallback)

        storable = metric.store_value(result, step=_logger_call.original_call.call_id.call_number)

        if storable:
            metric.store(_logger_output, key="metric")



import re

from typing import Type

from pydantic.main import BaseModel

from pypads.importext.semver import parse_constraint
from pypads.model.metadata import ModelObject
from pypads.model.models import LibSelectorModel
from pypads.utils.util import is_package_available, find_package_version, find_package_regex_versions


class VersionNotFoundException(Exception):
    pass


class LibSelector(ModelObject):
    """
    Selector class holding version constraint and name of a library. @see poetry sem versioning
    """

    @classmethod
    def get_model_cls(cls) -> Type[BaseModel]:
        return LibSelectorModel

    def __init__(self, *args, name, regex=False, constraint: str = "*", specificity: int = None, **kwargs) -> None:
        super().__init__(*args, name=name, regex=regex, constraint=constraint,
                         specificity=specificity or self._calc_specificity(), **kwargs)
        self._parsed_constraint = parse_constraint(constraint)

    @staticmethod
    def from_dict(library):
        if library is None:
            return None
        return LibSelector(name=library["name"], constraint=library["version"])

    def _calc_specificity(self):
        """
        Calculates a value how specific the selector is. The more specific it is the higher the value is.
        TODO do some magic here.
        :return:
        """
        return 0

    def is_installed(self):
        """
        Check if a match is installed
        :return:
        """
        if self.regex:
            return any({self.allows(version) for version in find_package_regex_versions(self.name).values() if
                        version is not None})
        else:
            if is_package_available(self.name):
                version = find_package_version(self.name)
                if version is None:
                    raise VersionNotFoundException("Couldn't find version for lib {}".format(self.name))
                return self.allows(version)
            return False

    def allows_any(self, other):  # type: (LibSelector) -> bool
        """
        Check if the constraint overlaps with another constaint.
        :param other:
        :return:
        """
        return re.compile(self.name).match(other.name) and self._parsed_constraint.allows_any(other._parsed_constraint)

    def allows(self, version):  # type: ("Version") -> bool
        """
        Check if the constraint allows given version number.
        :param version:
        :return:
        """
        from pypads.importext.semver import Version
        return self._parsed_constraint.allows(Version.parse(version))

    def __str__(self):
        return "LibSelector[name=" + self.name + "," + self.constraint + "]"



import types
from contextlib import contextmanager
from functools import wraps
from typing import Set

from app.env import InjectionLoggerEnv
from pypads import logger
from pypads.app.call import FunctionReference, CallAccessor, Call
from pypads.importext.mappings import MatchedMapping
from pypads.importext.wrapping.base_wrapper import BaseWrapper, Context
from pypads.injections.analysis.call_tracker import add_call, finish_call
from pypads.utils.util import get_experiment_id, get_run_id

error = False


class FunctionWrapper(BaseWrapper):

    def wrap(self, fn, context: Context, matched_mappings: Set[MatchedMapping]):
        # Only wrap functions not starting with "__"
        if (fn.__name__.startswith("__") or fn.__name__.startswith("_pypads")) and fn.__name__ is not "__init__":
            return fn

        if not context.has_original(fn) or not context.defined_stored_original(fn):
            for matched_mapping in matched_mappings:
                context.store_wrap_meta(matched_mapping, fn)

            context.store_original(fn)
            if context.is_class():
                return self._wrap_on_class(fn, context, matched_mappings)
            elif hasattr(context.container, fn.__name__):
                return self._wrap_on_object(fn, context, matched_mappings)
            else:
                logger.warning(str(
                    context) + " is no class and doesn't provide attribute with name " + str(
                    fn.__name__) + ". Couldn't access " + str(
                    fn) + " on it.")

    def _wrap_on_object(self, fn, context: Context, mappings: Set[MatchedMapping]):
        # Add module of class to the changed modules
        if hasattr(context.container, "__module__"):
            self._pypads.add_punched_module_name(context.container.__module__)

        # If we are already punched get original function instead of punched function
        if context.has_original(fn):
            fn = context.original(fn)
        else:
            fn = getattr(context.container, fn.__name__)

        # Get and add hooks
        hooks = self._get_hooked_fns(mappings)
        if len(hooks) > 0:
            fn_reference = FunctionReference(context, fn)
            return self.wrap_method_helper(fn_reference=fn_reference, hooks=hooks, mappings=mappings)
        else:
            return fn

    def _wrap_on_class(self, fn, context: Context, mappings: Set[MatchedMapping]):
        fn_name = fn.__name__

        # Find the real defining class
        defining_class = context.real_context(fn.__name__)

        # If there is no defining class we can't wrap on class
        if not defining_class:
            logger.warning("Defining class is None on context '" + str(context) + "' for fn '" + str(
                fn) + "'. Class punching might be lost in newly spawned processes.")
        else:
            # Add the module to the list of modules which where changed
            if isinstance(defining_class, Context):
                if hasattr(defining_class.container, "__module__"):
                    self._pypads.wrap_manager.module_wrapper.add_punched_module_name(
                        defining_class.container.__module__)

                # Get real fn from defining class
                fn = None
                try:
                    fn = getattr(defining_class.container, fn_name)
                except Exception as e:
                    context.real_context(fn.__name__)
                    logger.warning("Defining class doesn't define our function. Extraction failed: " + str(e))
            else:
                # skip broken context references
                logger.warning("Mapping an object which couldn't be contextualized."
                               " Context class: {} Mapping: {}".format(defining_class,
                                                                       ", ".join([str(m) for m in mappings])))
                return fn

        # skip wrong extractions
        if not fn or not callable(fn):
            return fn

        # if we have a property we should use fget instead
        if isinstance(fn, property):
            fn = fn.fget

        hooks = self._get_hooked_fns(mappings)
        if len(hooks) > 0:
            fn_reference = FunctionReference(context, fn)
            return self.wrap_method_helper(fn_reference=fn_reference, hooks=hooks, mappings=mappings)

    @contextmanager
    def _make_call(self, instance, fn_reference):
        accessor = CallAccessor.from_function_reference(fn_reference, instance)

        current_call = None
        call = None
        try:
            current_call: Call = self._pypads.call_tracker.current_call()
            if current_call and accessor.is_call_identity(current_call.call_id):
                call = current_call
            else:
                call = add_call(accessor)
            yield call
        finally:
            if call and not current_call == call:
                # print("c:" + str(call))
                # print("cc:" + str(current_call))
                finish_call(call)

    def wrap_method_helper(self, fn_reference: FunctionReference, hooks, mappings: Set[MatchedMapping]):
        """
        Helper to differentiate between functions, classmethods, static methods and wrap them
        :param fn_reference:
        :param hooks:
        :param mappings:
        :return:
        """
        fn = fn_reference.wrappee

        if fn_reference.is_static_method():
            @wraps(fn)
            def entry(*args, _pypads_hooks=hooks, _pypads_mapped_by=mappings, **kwargs):
                logger.debug("Call to tracked static method or function " + str(fn))

                global error
                if self._pypads.api.active_run():
                    error = False
                    with self._make_call(None, fn_reference) as call:
                        accessor = call.call_id
                        callback = fn

                        # for every hook add
                        if self._is_skip_recursion(accessor):
                            logger.info("Skipping " + str(accessor.context.container.__name__) + "." + str(
                                accessor.wrappee.__name__))
                            out = callback(*args, **kwargs)
                            return out

                        for (h, params) in hooks:
                            c = self._add_hook(h, params, callback, call, mappings)
                            if c:
                                callback = c

                        # start executing the stack
                        out = callback(*args, **kwargs)
                else:
                    if not error:
                        error = True
                        logger.error(
                            "No run was active to log your hooks. You may want to start a run with PyPads().start_track()")
                    out = fn(*args, **kwargs)
                return out
        elif fn_reference.is_function():
            @wraps(fn)
            def entry(_self, *args, _pypads_hooks=hooks, _pypads_mapped_by=mappings, **kwargs):
                # print("Call to tracked class method " + str(fn) + str(id(fn)))
                logger.debug("Call to tracked method " + str(fn))

                global error
                if self._pypads.api.active_run():
                    error = False
                    with self._make_call(_self, fn_reference) as call:
                        accessor = call.call_id
                        # add the function to the callback stack
                        callback = types.MethodType(fn, _self)

                        # for every hook add
                        if self._is_skip_recursion(accessor):
                            logger.info("Skipping " + str(accessor.context.container.__name__) + "." + str(
                                accessor.wrappee.__name__))
                            out = callback(*args, **kwargs)
                            return out

                        for (h, params) in hooks:
                            c = self._add_hook(h, params, callback, call, mappings)
                            if c:
                                callback = types.MethodType(c, _self)

                        # start executing the stack
                        out = callback(*args, **kwargs)
                else:
                    if not error:
                        error = True
                        logger.error(
                            "No run was active to log your hooks. You may want to start a run with PyPads().start_track()")
                    callback = types.MethodType(fn, _self)
                    out = callback(*args, **kwargs)
                return out

        elif fn_reference.is_class_method():
            @wraps(fn)
            def entry(_cls, *args, _pypads_hooks=hooks, _pypads_mapped_by=mappings, **kwargs):
                logger.debug("Call to tracked class method " + str(fn))

                global error
                if self._pypads.api.active_run():
                    error = False
                    with self._make_call(_cls, fn_reference) as call:
                        accessor = call.call_id
                        # add the function to the callback stack
                        callback = types.MethodType(fn, _cls)

                        # for every hook add
                        if self._is_skip_recursion(accessor):
                            logger.info("Skipping " + str(accessor.context.container.__name__) + "." + str(
                                accessor.wrappee.__name__))
                            out = callback(*args, **kwargs)
                            return out

                        for (h, params) in hooks:
                            c = self._add_hook(h, params, callback, call, mappings)
                            if c:
                                callback = types.MethodType(c, _cls)

                        # start executing the stack
                        out = callback(*args, **kwargs)
                else:
                    if not error:
                        error = True
                        logger.error(
                            "No run was active to log your hooks. You may want to start a run with PyPads().start_track()")
                    callback = types.MethodType(fn, _cls)
                    out = callback(*args, **kwargs)
                return out

        elif fn_reference.is_wrapped():
            tmp_fn = getattr(fn_reference.context.container, fn.__name__)

            @wraps(tmp_fn)
            def entry(_self, *args, _pypads_hooks=hooks, _pypads_mapped_by=mappings, **kwargs):
                logger.debug("Call to tracked _IffHasAttrDescriptor " + str(fn))

                global error
                if self._pypads.api.active_run():
                    error = False
                    with self._make_call(_self, fn_reference) as call:
                        accessor = call.call_id
                        # add the function to the callback stack
                        callback = fn.__get__(_self)

                        # for every hook add
                        if self._is_skip_recursion(accessor):
                            logger.info("Skipping " + str(accessor.context.container.__name__) + "." + str(
                                accessor.wrappee.__name__))
                            out = callback(*args, **kwargs)
                            return out

                        for (h, params) in hooks:
                            c = self._add_hook(h, params, callback, call, mappings)
                            if c:
                                callback = types.MethodType(c, _self)

                        # start executing the stack
                        out = callback(*args, **kwargs)
                else:
                    if not error:
                        error = True
                        logger.error(
                            "No run was active to log your hooks. You may want to start a run with PyPads().start_track()")

                    callback = fn.__get__(_self)
                    out = callback(*args, **kwargs)
                return out
        else:
            return fn
        fn_reference.context.overwrite(fn.__name__, entry)
        # print("Wrapped " + str(fn) + str(id(fn)))
        return entry

    def _add_hook(self, hook, params, callback, call: Call, mappings):
        # For every hook we defined on the given function in out mapping file execute it before running the code
        if not call.has_hook(hook):
            return self._get_env_setter(
                _pypads_env=InjectionLoggerEnv(mappings, hook, callback, call, params, get_experiment_id(),
                                               get_run_id()))
        else:
            logger.debug(str(hook) + " is tracked multiple times on " + str(call) + ". Ignoring second hooking.")
            return None

    def _get_env_setter(self, _pypads_env: InjectionLoggerEnv):
        """
        This wrapper sets the context
        :return:
        """
        env = _pypads_env
        call: Call = env.call
        cid = call.call_id

        # TODO maybe use https://gist.github.com/MacHu-GWU/0170849f693aa5f8d129aa03fc358305
        if cid.is_static_method():
            @wraps(cid.wrappee)
            def env_setter(*args, _pypads_env=env, **kwargs):
                logger.debug("Static method hook " + str(cid.context) + str(cid.wrappee) + str(env.hook))
                return self._wrapped_inner_function(None, *args, _pypads_env=_pypads_env, **kwargs)

            return env_setter
        elif cid.is_function():
            @wraps(cid.wrappee)
            def env_setter(_self, *args, _pypads_env=env, **kwargs):
                logger.debug("Method hook " + str(cid.context) + str(cid.wrappee) + str(env.hook))
                return self._wrapped_inner_function(_self, *args, _pypads_env=_pypads_env, **kwargs)

            return env_setter
        elif cid.is_class_method():
            @wraps(cid.wrappee)
            def env_setter(_cls, *args, _pypads_env=env, **kwargs):
                logger.debug("Class method hook " + str(cid.context) + str(cid.wrappee) + str(env.hook))
                return self._wrapped_inner_function(_cls, *args, _pypads_env=_pypads_env,
                                                    **kwargs)

            return env_setter
        elif cid.is_wrapped():
            tmp_fn = getattr(cid.context.container, cid.wrappee.__name__)

            @wraps(tmp_fn)
            def env_setter(_self, *args, _pypads_env=env, **kwargs):
                logger.debug("Method hook " + str(cid.context) + str(cid.wrappee) + str(env.hook))
                return self._wrapped_inner_function(_self, *args, _pypads_env=_pypads_env, **kwargs)

            return env_setter
        else:
            raise ValueError("Failed!")

    @staticmethod
    def _wrapped_inner_function(_self, *args, _pypads_env: InjectionLoggerEnv, **kwargs):
        """
        Wrapped function logic. This executes all hooks and the function itself.
        :param _self: Reference to
        :param args:
        :param kwargs:
        :return:
        """

        env = _pypads_env
        call = env.call
        if not call.has_hook(env.hook):
            call.add_hook(env.hook)

            try:
                # check for name collision in parameters
                if set([k for k, v in kwargs.items()]) & set(
                        [k for k, v in env.parameter.items()]):
                    logger.warning("Hook parameter is overwriting a parameter in the standard "
                                   "model call. This most likely will produce side effects.")

                if env.hook:
                    return env.hook(_self, _pypads_env=_pypads_env, *args, **kwargs)
                return env.callback(*args, **kwargs)
            finally:
                call.remove_hook(env.hook)
        else:
            return env.callback(*args, **kwargs)

    def _is_skip_recursion(self, accessor):
        try:
            config = self._pypads.config

            if 'recursion_depth' in config and config['recursion_depth'] is not -1:
                if self._pypads.call_tracker.call_depth() > config['recursion_depth'] + 1:
                    return True
            if 'recursion_identity' in config and config['recursion_identity']:
                if self._pypads.call_tracker.has_call_identity(accessor):
                    return True
            return False
        except Exception as e:
            return False
